(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('react')) :
	typeof define === 'function' && define.amd ? define(['react'], factory) :
	(global.Ligand = factory(global.React));
}(this, (function (React) { 'use strict';

var React__default = 'default' in React ? React['default'] : React;

// returns a fragment
var readMol = (function (f) {
  f = f.split(/\r?\n/);

  var countsLine = f[3];

  var numberOfAtoms = Number(countsLine.substring(0, 3));
  var numberOfBonds = Number(countsLine.substring(3, 6));

  var fragment = {
    nodes: [],
    bonds: []
  };

  // reading the atom block
  for (var i = 4; i < 4 + numberOfAtoms; ++i) {
    var atomLine = f[i];

    var node = {
      x: Math.floor(Number(atomLine.substring(0, 10)) * 50 + 49999),
      y: Math.floor(-Number(atomLine.substring(10, 20)) * 50 + 49999),
      z: Math.floor(Number(atomLine.substring(20, 30)) * 50 + 49999),
      label: atomLine.substring(31, 34).trim() + 'H₂₂',
      focusCharacterRange: [0, atomLine.substring(31, 34).trim().length], //[1, 1+atomLine.substring(31,34).trim().length], // one range per line
      focusLine: 0, // assuming one line
      beingEdited: false,
      latestMouseDownXY: {
        x: undefined,
        y: undefined
      }
    };

    fragment.nodes.push(node);
  }

  // reading the bond block
  for (var _i = 4 + numberOfAtoms; _i < 4 + numberOfAtoms + numberOfBonds; ++_i) {
    var bondLine = f[_i];

    var bond = {
      firstNodeIndex: Number(bondLine.substring(0, 3)) - 1,
      secondNodeIndex: Number(bondLine.substring(3, 6)) - 1,
      type: Number(bondLine.substring(6, 9)),
      stereo: Number(bondLine.substring(9, 12))
    };

    fragment.bonds.push(bond);
  }

  return fragment;
});

/*
TO GET THE ARRAY, GO TO:
https://en.wikipedia.org/wiki/List_of_chemical_elements
AND RUN THIS IN THE CONSOLE:
JSON.stringify(Array.prototype.reduce.call(document.querySelector('#collapsibleTable0 > tbody').getElementsByTagName('tr'), (acc, tr) => {if (tr.getElementsByTagName('td').length < 2) return acc; acc.push({symbol: tr.getElementsByTagName('td')[1].textContent, name: tr.getElementsByTagName('td')[2].textContent}); return acc;}, [{symbol: 'OOPS', name: 'Zeroth element'}]), null, 2)
*/

var ELEMENTS = [{
  "symbol": "OOPS",
  "name": "Zeroth element"
}, {
  "symbol": "H",
  "name": "Hydrogen"
}, {
  "symbol": "He",
  "name": "Helium"
}, {
  "symbol": "Li",
  "name": "Lithium"
}, {
  "symbol": "Be",
  "name": "Beryllium"
}, {
  "symbol": "B",
  "name": "Boron"
}, {
  "symbol": "C",
  "name": "Carbon"
}, {
  "symbol": "N",
  "name": "Nitrogen"
}, {
  "symbol": "O",
  "name": "Oxygen"
}, {
  "symbol": "F",
  "name": "Fluorine"
}, {
  "symbol": "Ne",
  "name": "Neon"
}, {
  "symbol": "Na",
  "name": "Sodium"
}, {
  "symbol": "Mg",
  "name": "Magnesium"
}, {
  "symbol": "Al",
  "name": "Aluminium"
}, {
  "symbol": "Si",
  "name": "Silicon"
}, {
  "symbol": "P",
  "name": "Phosphorus"
}, {
  "symbol": "S",
  "name": "Sulfur"
}, {
  "symbol": "Cl",
  "name": "Chlorine"
}, {
  "symbol": "Ar",
  "name": "Argon"
}, {
  "symbol": "K",
  "name": "Potassium"
}, {
  "symbol": "Ca",
  "name": "Calcium"
}, {
  "symbol": "Sc",
  "name": "Scandium"
}, {
  "symbol": "Ti",
  "name": "Titanium"
}, {
  "symbol": "V",
  "name": "Vanadium"
}, {
  "symbol": "Cr",
  "name": "Chromium"
}, {
  "symbol": "Mn",
  "name": "Manganese"
}, {
  "symbol": "Fe",
  "name": "Iron"
}, {
  "symbol": "Co",
  "name": "Cobalt"
}, {
  "symbol": "Ni",
  "name": "Nickel"
}, {
  "symbol": "Cu",
  "name": "Copper"
}, {
  "symbol": "Zn",
  "name": "Zinc"
}, {
  "symbol": "Ga",
  "name": "Gallium"
}, {
  "symbol": "Ge",
  "name": "Germanium"
}, {
  "symbol": "As",
  "name": "Arsenic"
}, {
  "symbol": "Se",
  "name": "Selenium"
}, {
  "symbol": "Br",
  "name": "Bromine"
}, {
  "symbol": "Kr",
  "name": "Krypton"
}, {
  "symbol": "Rb",
  "name": "Rubidium"
}, {
  "symbol": "Sr",
  "name": "Strontium"
}, {
  "symbol": "Y",
  "name": "Yttrium"
}, {
  "symbol": "Zr",
  "name": "Zirconium"
}, {
  "symbol": "Nb",
  "name": "Niobium"
}, {
  "symbol": "Mo",
  "name": "Molybdenum"
}, {
  "symbol": "Tc",
  "name": "Technetium"
}, {
  "symbol": "Ru",
  "name": "Ruthenium"
}, {
  "symbol": "Rh",
  "name": "Rhodium"
}, {
  "symbol": "Pd",
  "name": "Palladium"
}, {
  "symbol": "Ag",
  "name": "Silver"
}, {
  "symbol": "Cd",
  "name": "Cadmium"
}, {
  "symbol": "In",
  "name": "Indium"
}, {
  "symbol": "Sn",
  "name": "Tin"
}, {
  "symbol": "Sb",
  "name": "Antimony"
}, {
  "symbol": "Te",
  "name": "Tellurium"
}, {
  "symbol": "I",
  "name": "Iodine"
}, {
  "symbol": "Xe",
  "name": "Xenon"
}, {
  "symbol": "Cs",
  "name": "Caesium"
}, {
  "symbol": "Ba",
  "name": "Barium"
}, {
  "symbol": "La",
  "name": "Lanthanum"
}, {
  "symbol": "Ce",
  "name": "Cerium"
}, {
  "symbol": "Pr",
  "name": "Praseodymium"
}, {
  "symbol": "Nd",
  "name": "Neodymium"
}, {
  "symbol": "Pm",
  "name": "Promethium"
}, {
  "symbol": "Sm",
  "name": "Samarium"
}, {
  "symbol": "Eu",
  "name": "Europium"
}, {
  "symbol": "Gd",
  "name": "Gadolinium"
}, {
  "symbol": "Tb",
  "name": "Terbium"
}, {
  "symbol": "Dy",
  "name": "Dysprosium"
}, {
  "symbol": "Ho",
  "name": "Holmium"
}, {
  "symbol": "Er",
  "name": "Erbium"
}, {
  "symbol": "Tm",
  "name": "Thulium"
}, {
  "symbol": "Yb",
  "name": "Ytterbium"
}, {
  "symbol": "Lu",
  "name": "Lutetium"
}, {
  "symbol": "Hf",
  "name": "Hafnium"
}, {
  "symbol": "Ta",
  "name": "Tantalum"
}, {
  "symbol": "W",
  "name": "Tungsten"
}, {
  "symbol": "Re",
  "name": "Rhenium"
}, {
  "symbol": "Os",
  "name": "Osmium"
}, {
  "symbol": "Ir",
  "name": "Iridium"
}, {
  "symbol": "Pt",
  "name": "Platinum"
}, {
  "symbol": "Au",
  "name": "Gold"
}, {
  "symbol": "Hg",
  "name": "Mercury"
}, {
  "symbol": "Tl",
  "name": "Thallium"
}, {
  "symbol": "Pb",
  "name": "Lead"
}, {
  "symbol": "Bi",
  "name": "Bismuth"
}, {
  "symbol": "Po",
  "name": "Polonium"
}, {
  "symbol": "At",
  "name": "Astatine"
}, {
  "symbol": "Rn",
  "name": "Radon"
}, {
  "symbol": "Fr",
  "name": "Francium"
}, {
  "symbol": "Ra",
  "name": "Radium"
}, {
  "symbol": "Ac",
  "name": "Actinium"
}, {
  "symbol": "Th",
  "name": "Thorium"
}, {
  "symbol": "Pa",
  "name": "Protactinium"
}, {
  "symbol": "U",
  "name": "Uranium"
}, {
  "symbol": "Np",
  "name": "Neptunium"
}, {
  "symbol": "Pu",
  "name": "Plutonium"
}, {
  "symbol": "Am",
  "name": "Americium"
}, {
  "symbol": "Cm",
  "name": "Curium"
}, {
  "symbol": "Bk",
  "name": "Berkelium"
}, {
  "symbol": "Cf",
  "name": "Californium"
}, {
  "symbol": "Es",
  "name": "Einsteinium"
}, {
  "symbol": "Fm",
  "name": "Fermium"
}, {
  "symbol": "Md",
  "name": "Mendelevium"
}, {
  "symbol": "No",
  "name": "Nobelium"
}, {
  "symbol": "Lr",
  "name": "Lawrencium"
}, {
  "symbol": "Rf",
  "name": "Rutherfordium"
}, {
  "symbol": "Db",
  "name": "Dubnium"
}, {
  "symbol": "Sg",
  "name": "Seaborgium"
}, {
  "symbol": "Bh",
  "name": "Bohrium"
}, {
  "symbol": "Hs",
  "name": "Hassium"
}, {
  "symbol": "Mt",
  "name": "Meitnerium"
}, {
  "symbol": "Ds",
  "name": "Darmstadtium"
}, {
  "symbol": "Rg",
  "name": "Roentgenium"
}, {
  "symbol": "Cn",
  "name": "Copernicium"
}, {
  "symbol": "Nh",
  "name": "Nihonium"
}, {
  "symbol": "Fl",
  "name": "Flerovium"
}, {
  "symbol": "Mc",
  "name": "Moscovium"
}, {
  "symbol": "Lv",
  "name": "Livermorium"
}, {
  "symbol": "Ts",
  "name": "Tennessine"
}, {
  "symbol": "Og",
  "name": "Oganesson"
}];

var CONSTANTS = {
  FONT_SIZE: 24,
  BOND_BAR_WIDTH: 2.5,
  LABEL_BOND_SPACING: 8,
  ADDED_MARGIN_FOR_CANVASES: 20,
  SVG_PADDING: 20
};

function createPoint(x, y) {
  return { x: x, y: y };
}

function addPoints(p1, p2) {
  return {
    x: p1.x + p2.x,
    y: p1.y + p2.y
  };
}

function subtractPoints(p1, p2) {
  return {
    x: p1.x - p2.x,
    y: p1.y - p2.y
  };
}

function scalePoint(p, k) {
  return {
    x: k * p.x,
    y: k * p.y
  };
}

function slidePoint(p1, p2, t) {
  return createPoint(p1.x + t * (p2.x - p1.x), p1.y + t * (p2.y - p1.y));
}

function pointToPathString(p) {
  return p.x + ' ' + p.y;
}

function projectAPointOntoALine(A, B, C) {
  // A and B are the point that define the line.
  // C is the point that we a projecting onto the line.
  // AB is the vector from A to B.
  // ABperp is the AB vector rotated by 90 degrees counterclockwise.
  // u is the slider for AB vector if we attach it to A.
  // v is the slider for ABperp vector if we attach it to C.
  // u and v determine the coordinates of the projection point.
  // P is the projection point.

  var AB = {
    x: B.x - A.x,
    y: B.y - A.y
  };
  var ABperp = {
    x: A.y - B.y,
    y: B.x - A.x
  };

  // A + u * AB == C + v * ABperp.

  // A.x + u * AB.x == C.x + v * ABperp.x
  // A.y + u * AB.y == C.y + v * ABperp.y

  // u == (C.x + v * ABperp.x - A.x) / AB.x
  // u == (C.y + v * ABperp.y - A.y) / AB.y

  // (C.x + v * ABperp.x - A.x) / AB.x == (C.y + v * ABperp.y - A.y) / AB.y

  // (C.x + v * ABperp.x - A.x) * AB.y == (C.y + v * ABperp.y - A.y) * AB.x

  // C.x * AB.y + v * ABperp.x * AB.y - A.x * AB.y == C.y * AB.x + v * ABperp.y * AB.x - A.y * AB.x

  // v * ABperp.x * AB.y - v * ABperp.y * AB.x == C.y * AB.x - A.y * AB.x - C.x * AB.y + A.x * AB.y

  // v = (C.y * AB.x - A.y * AB.x - C.x * AB.y + A.x * AB.y) / (ABperp.x * AB.y - ABperp.y * AB.x)

  var v = (C.y * AB.x - A.y * AB.x - C.x * AB.y + A.x * AB.y) / (ABperp.x * AB.y - ABperp.y * AB.x);
  var u = AB.x ? (C.x + v * ABperp.x - A.x) / AB.x : (C.y + v * ABperp.y - A.y) / AB.y;
  var P = {
    x: A.x + u * AB.x,
    y: A.y + u * AB.y
  };
  return { u: u, v: v, P: P };
}

function distanceBetweenTwoPoints(A, B) {
  var d = subtractPoints(B, A);
  return Math.sqrt(d.x * d.x + d.y * d.y);
}

function getBondsInvolved(nodeIndex, fragmentBonds) {
  return fragmentBonds.filter(function (bond) {
    return bond.firstNodeIndex === nodeIndex || bond.secondNodeIndex === nodeIndex;
  });
}

function getBondD(bond, fragmentNodes) {
  var firstNode = fragmentNodes[bond.firstNodeIndex];
  var secondNode = fragmentNodes[bond.secondNodeIndex];
  return subtractPoints(secondNode, firstNode);
}

function getVectorLength(d) {
  return Math.sqrt(d.x * d.x + d.y * d.y);
}

function getVectorAngle(d) {
  return Math.atan2(d.y, d.x);
}

function reversePoint(p) {
  return {
    x: -p.x,
    y: -p.y
  };
}

// t is the proportion of segment1 to the intersection point
function intersection(segment1FirstPointCoordinates, segment1SecondPointCoordinates, //
segment2FirstPointCoordinates, segment2SecondPointCoordinates //
) {

  var p = segment1FirstPointCoordinates;
  var q = segment2FirstPointCoordinates;

  var r = {
    x: segment1SecondPointCoordinates.x - segment1FirstPointCoordinates.x,
    y: segment1SecondPointCoordinates.y - segment1FirstPointCoordinates.y
  };
  var s = {
    x: segment2SecondPointCoordinates.x - segment2FirstPointCoordinates.x,
    y: segment2SecondPointCoordinates.y - segment2FirstPointCoordinates.y
  };
  //points p and q
  //vectors r and s from those points

  //intersection point is at (p + t*r) or (q + u*s)
  //p.x + t*r.x = q.x + u*s.x
  //p.y + t*r.y = q.y + u*s.y <=> u = (p.y + t*r.y - q.y) / s.y
  //
  // =>
  //
  //p.x + t*r.x = q.x + [(p.y + t*r.y - q.y) / s.y * s.x] = q.x + p.y*s.x/s.y + t*r.y*s.x/s.y - q.y*s.x/s.y
  // <=>
  //t = (q.x + p.y*s.x/s.y - q.y*s.x/s.y - p.x) / (r.x - r.y*s.x/s.y)

  if (Math.abs(s.y) < 0.0001) {
    //segment 2 is horizontal
    return ((segment2SecondPointCoordinates.y + segment2FirstPointCoordinates.y) / 2 - segment1FirstPointCoordinates.y) / r.y;
  } else {
    return (q.x + p.y * s.x / s.y - q.y * s.x / s.y - p.x) / (r.x - r.y * s.x / s.y);
  }
}

function cutBondsWithLabel(nodeIndex, filledPixelsData, fragmentNodes, fragmentBonds) {
  var bondsInvolved = getBondsInvolved(nodeIndex, fragmentBonds);

  bondsInvolved.forEach(function (bond) {

    var firstNode = fragmentNodes[bond.firstNodeIndex];
    var secondNode = fragmentNodes[bond.secondNodeIndex];

    var d = getBondD(bond, fragmentNodes);
    var length = getVectorLength(d);
    // If horizontal unit vector is along X, vertical unit vector is directed upwards.
    var bondVerticalUnitVector = createPoint(0 - d.y / length, d.x / length);
    var tStep = 1 / length / 2;

    if (bond.firstNodeIndex === nodeIndex) {

      switch (bond.type) {
        case 1:
        default:

          bond.firstT = 0;

          for (var filledPixelIndex = 0; filledPixelIndex < filledPixelsData.length; filledPixelIndex += 2) {

            var C = {
              x: filledPixelsData[filledPixelIndex],
              y: filledPixelsData[filledPixelIndex + 1]
            };

            var _projectAPointOntoALi = projectAPointOntoALine({ x: 0, y: 0 }, subtractPoints(secondNode, firstNode), C),
                P = _projectAPointOntoALi.P,
                u = _projectAPointOntoALi.u;

            if (distanceBetweenTwoPoints(P, C) < CONSTANTS.BOND_BAR_WIDTH / 2) {
              bond.firstT = Math.max(bond.firstT, u);
            }
          }

          // bond.firstT = 0;
          //
          // var t = 0;
          // for (t = 0; t <= 1; t+=tStep) {
          //
          //
          //
          //
          //   if (!polygonOverlapsImg([
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), firstNode),
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), firstNode),
          //     subtractPoints(
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       firstNode
          //     ),
          //     subtractPoints(
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       firstNode
          //     ),
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), firstNode)
          //   ], filledPixelsData)) break;
          //
          // }
          //
          // bond.firstT = t;

          break;

        // case 2:

        // var t = 0;
        // for (t = 0; t <= 1; t+=0.005) {
        //
        //
        //
        //
        //   if (!polygonOverlapsImg(ctx, [
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //       secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //       t
        //     ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //       secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //       t
        //     ).added(bondHorizontalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //       secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //       t
        //     ).added(bondHorizontalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //       secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //       t
        //     ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
        //   ], canvas, thisAtom)
        //
        //   &&
        //
        //   !polygonOverlapsImg(ctx, [
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //       secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //       t
        //     ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //       secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //       t
        //     ).added(bondHorizontalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //       secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //       t
        //     ).added(bondHorizontalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //       secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //       t
        //     ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
        //   ], canvas, thisAtom)) break;
        //
        // }
        //
        // bond.firstT = t;
        //
        // break;

      }
    } else {
      //
      //
      //
      switch (bond.type) {
        case 1:
        default:

          bond.secondT = 1;

          for (var _filledPixelIndex = 0; _filledPixelIndex < filledPixelsData.length; _filledPixelIndex += 2) {

            var _C = {
              x: filledPixelsData[_filledPixelIndex],
              y: filledPixelsData[_filledPixelIndex + 1]
            };

            var _projectAPointOntoALi2 = projectAPointOntoALine(subtractPoints(firstNode, secondNode), { x: 0, y: 0 }, _C),
                P = _projectAPointOntoALi2.P,
                u = _projectAPointOntoALi2.u;

            if (distanceBetweenTwoPoints(P, _C) < CONSTANTS.BOND_BAR_WIDTH / 2) {
              bond.secondT = Math.min(bond.secondT, u);
            }
          }
          //
          // bond.secondT = 1;
          // var t = 1;
          // for (t = 1; t >= 0; t-=tStep) {
          //
          //
          //   // if (!polygonOverlapsImg(ctx, [
          //   //   firstNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   firstNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   slide(
          //   //     firstNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     secondNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     t
          //   //   ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   slide(
          //   //     firstNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     secondNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     t
          //   //   ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   slide(
          //   //     firstNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     secondNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     t
          //   //   ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   slide(
          //   //     firstNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     secondNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     t
          //   //   ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
          //   // ], canvas, thisAtom)) break;
          //
          //   if (!polygonOverlapsImg([
          //     subtractPoints(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       secondNode
          //     ),
          //     subtractPoints(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       secondNode
          //     ),
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), secondNode),
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), secondNode),
          //     subtractPoints(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       secondNode
          //     )
          //   ], filledPixelsData)) break;
          //
          // }
          //
          // bond.secondT = t;

          break;
        //   case 2:
        //   var t = 1;
        //   for (t = 1; t >= 0; t-=0.005) {
        //
        //
        //     if (!polygonOverlapsImg(ctx, [
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //         secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //         t
        //       ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //         secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //         t
        //       ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //         secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //         t
        //       ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //         secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //         t
        //       ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
        //     ], canvas, thisAtom)
        //
        //     &&
        //
        //     !polygonOverlapsImg(ctx, [
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //         secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //         t
        //       ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //         secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //         t
        //       ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //         secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //         t
        //       ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //         secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //         t
        //       ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
        //     ], canvas, thisAtom)) break;
        //
        //   }
        //
        //   bond.secondT = t;
        //
        //   break;
      }
      //
      //
      //
      // //barCoordinatesArray.secondUpT = t;
      // //barCoordinatesArray.secondDownT = t;
    }
  });
}

function cutBondsWithoutLabel(nodeIndex, fragmentNodes, fragmentBonds) {

  var bondsInvolved = getBondsInvolved(nodeIndex, fragmentBonds);

  bondsInvolved.sort(function (bond1, bond2) {

    var d1 = getBondD(bond1, fragmentNodes);
    var d2 = getBondD(bond2, fragmentNodes);

    var bond1IsDirectedOutwards = bond1.firstNodeIndex === nodeIndex;
    var bond2IsDirectedOutwards = bond2.firstNodeIndex === nodeIndex;

    var angle1 = void 0,
        angle2 = void 0;
    if (bond1IsDirectedOutwards) {
      if (bond2IsDirectedOutwards) {
        //bond 1 and bond 2 are directed outwards
        angle1 = getVectorAngle(d1);
        angle2 = getVectorAngle(d2);
      } else {
        //bond 1 is directed outwards and bond 2 is directed inwards
        angle1 = getVectorAngle(d1);
        angle2 = getVectorAngle(reversePoint(d2));
      }
    } else {
      if (bond2IsDirectedOutwards) {
        //bond 1 is directed inwards and bond 2 is directed outwards
        angle1 = getVectorAngle(reversePoint(d1));
        angle2 = getVectorAngle(d2);
      } else {
        //bond 1 and bond 2 are directed inwards
        angle1 = getVectorAngle(reversePoint(d1));
        angle2 = getVectorAngle(reversePoint(d2));
      }
    }

    // bond1 should come first if angle1 < angle2

    return angle1 - angle2;
  });

  if (bondsInvolved.length === 1) return;

  for (var s = 0; s < bondsInvolved.length; s++) {

    var bond1 = bondsInvolved[s];
    var bond2 = bondsInvolved[(s + 1) % bondsInvolved.length];

    var bond1IsDirectedOutwards = bond1.firstNodeIndex === nodeIndex;
    var bond2IsDirectedOutwards = bond2.firstNodeIndex === nodeIndex;

    var d1 = getBondD(bond1, fragmentNodes);
    var length1 = getVectorLength(d1);
    var bond1VerticalUnitVector = createPoint(-d1.y / length1, d1.x / length1);

    var d2 = getBondD(bond2, fragmentNodes);
    var length2 = getVectorLength(d2);
    var bond2VerticalUnitVector = createPoint(-d2.y / length2, d2.x / length2);

    var bond1FirstNode = fragmentNodes[bond1.firstNodeIndex];
    var bond1SecondNode = fragmentNodes[bond1.secondNodeIndex];

    var bond2FirstNode = fragmentNodes[bond2.firstNodeIndex];
    var bond2SecondNode = fragmentNodes[bond2.secondNodeIndex];

    var bond1FirstUp = addPoints(bond1FirstNode, scalePoint(bond1VerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2));
    var bond1FirstDown = addPoints(bond1FirstNode, scalePoint(bond1VerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2));

    var bond1SecondUp = addPoints(bond1SecondNode, scalePoint(bond1VerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2));
    var bond1SecondDown = addPoints(bond1SecondNode, scalePoint(bond1VerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2));

    var bond2FirstUp = addPoints(bond2FirstNode, scalePoint(bond2VerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2));
    var bond2FirstDown = addPoints(bond2FirstNode, scalePoint(bond2VerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2));

    var bond2SecondUp = addPoints(bond2SecondNode, scalePoint(bond2VerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2));
    var bond2SecondDown = addPoints(bond2SecondNode, scalePoint(bond2VerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2));

    if (bond1IsDirectedOutwards) {
      if (bond2IsDirectedOutwards) {
        //bond 1 and bond 2 are directed outwards
        bond1.firstUpT = intersection(bond1FirstUp, bond1SecondUp, bond2FirstDown, bond2SecondDown);
        bond2.firstDownT = intersection(bond2FirstDown, bond2SecondDown, bond1FirstUp, bond1SecondUp);
      } else {
        //bond 1 is directed outwards and bond 2 is directed inwards
        bond1.firstUpT = intersection(bond1FirstUp, bond1SecondUp, bond2FirstUp, bond2SecondUp);
        bond2.secondUpT = intersection(bond2FirstUp, bond2SecondUp, bond1FirstUp, bond1SecondUp);
      }
    } else {
      if (bond2IsDirectedOutwards) {
        //bond 1 is directed inwards and bond 2 is directed outwards
        bond1.secondDownT = intersection(bond1FirstDown, bond1SecondDown, bond2FirstDown, bond2SecondDown);
        bond2.firstDownT = intersection(bond2FirstDown, bond2SecondDown, bond1FirstDown, bond1SecondDown);
      } else {
        //bond 1 and bond 2 are directed inwards
        bond1.secondDownT = intersection(bond1FirstDown, bond1SecondDown, bond2FirstUp, bond2SecondUp);
        bond2.secondUpT = intersection(bond2FirstUp, bond2SecondUp, bond1FirstDown, bond1SecondDown);
      }
    }
  }
}

function getCanvas(node, canvases) {
  var canvasCandidates = canvases.filter(function (c) {
    return c.label === node.label && c.focusCharacterRange[0] === node.focusCharacterRange[0] && c.focusCharacterRange[1] === node.focusCharacterRange[1] && c.focusLine === node.focusLine;
  });
  if (canvasCandidates.length) {
    return canvasCandidates[0];
  }
  return null;
}

function getPathOfTheObjectBeingPointedAt(mousePos, objects) {
  var pathOfTheObjectBeingHoveredOver = undefined;
  objects.fragments.forEach(function (fragment, fragmentIndex) {

    fragment.bonds.forEach(function (bond, bondIndex) {
      var firstNode = fragment.nodes[bond.firstNodeIndex];
      var secondNode = fragment.nodes[bond.secondNodeIndex];
      var bondLength = Math.sqrt(Math.pow(secondNode.x - firstNode.x, 2) + Math.pow(secondNode.y - firstNode.y, 2));
      if (Math.sqrt(Math.pow(mousePos.x - firstNode.x, 2) + Math.pow(mousePos.y - firstNode.y, 2)) + Math.sqrt(Math.pow(mousePos.x - secondNode.x, 2) + Math.pow(mousePos.y - secondNode.y, 2)) < bondLength + 5) {
        pathOfTheObjectBeingHoveredOver = ['fragments', fragmentIndex, 'bonds', bondIndex];
      }
    });

    fragment.nodes.forEach(function (node, nodeIndex) {
      if (Math.sqrt(Math.pow(mousePos.x - node.x, 2) + Math.pow(mousePos.y - node.y, 2)) < 5) {
        pathOfTheObjectBeingHoveredOver = ['fragments', fragmentIndex, 'nodes', nodeIndex];
      }
    });
  });
  return pathOfTheObjectBeingHoveredOver;
}

function arraysAreEqual(a1, a2) {
  if (a1.length === a2.length) {
    for (var i = 0; i < a1.length; ++i) {
      if (a1[i] !== a2[i]) return false;
    }
    return true;
  }
  return false;
}

function arrayOfArraysIncludesArray(arrayOfArays, a) {
  return arrayOfArays.some(function (b) {
    return arraysAreEqual(b, a);
  });
}

function getHaloFromObjectPath(objectPath, objectPathIndex, hoverIsTrueSelectionIsFalse, state) {
  // Assuming objects are either nodes or bonds
  // which also means that objectPath is an array of length 4
  switch (objectPath[0]) {
    case 'fragments':
      switch (objectPath[2]) {
        case 'nodes':
          var node = state.objects.fragments[objectPath[1]].nodes[objectPath[3]];
          return React__default.createElement('rect', {
            key: objectPathIndex,
            x: node.x - 10,
            y: node.y - 10,
            width: 20,
            height: 20,
            style: {
              fill: hoverIsTrueSelectionIsFalse ? 'rgba(0, 0, 255, 0.5)' : 'rgb(255, 100, 255)'
            }
          });
          break;
        case 'bonds':
          var bond = state.objects.fragments[objectPath[1]].bonds[objectPath[3]];
          var firstNode = state.objects.fragments[objectPath[1]].nodes[bond.firstNodeIndex];
          var secondNode = state.objects.fragments[objectPath[1]].nodes[bond.secondNodeIndex];
          return React__default.createElement('line', {
            key: objectPathIndex,
            x1: firstNode.x,
            y1: firstNode.y,
            x2: secondNode.x,
            y2: secondNode.y,
            style: {
              strokeWidth: '15',
              stroke: hoverIsTrueSelectionIsFalse ? 'rgba(0, 0, 255, 0.5)' : 'rgb(100, 255, 100)',
              strokeLinecap: 'round'
            }
          });
          break;
        default:
          throw 'hmmmm';
      }
      break;
    default:
      throw 'hmmmmmm';
  }
}

// The function mutates the fragment's nodes.
var positionFloatingHydrogens = function (fragment) {
  fragment.nodes.forEach(function (node, nodeIndex) {
    if (node.labelMode === 'CHEMICAL_ELEMENT_PLUS_FLOATING_HYDROGENS') {

      if (node.labelModeData.atomicNumber === 6) {
        // There is no label.

        // Editing the node.
        node.label = '';
        node.focusCharacterRange = [0, 0];
        node.focusLine = 0;
      } else {
        // There is a label.

        var hydrogensLabel = node.labelModeData.numberOfHydrogens ? 'H' + (node.labelModeData.numberOfHydrogens === 1 ? '' : node.labelModeData.numberOfHydrogens) : '';
        if (hydrogensLabel) {
          // We need to position the hydrogens.

          var leftBusiness = 0;
          var rightBusiness = 0;

          var bondsInvolved = getBondsInvolved(nodeIndex, fragment.bonds);

          bondsInvolved.forEach(function (bond) {
            var bondIsDirectedOutwards = bond.firstNodeIndex === nodeIndex;

            var firstNode = fragment.nodes[bond.firstNodeIndex];
            var secondNode = fragment.nodes[bond.secondNodeIndex];

            var dX = secondNode.x - firstNode.x;
            var dY = secondNode.y - firstNode.y;

            var angle = bondIsDirectedOutwards ? Math.atan2(dY, dX) : Math.atan2(-dY, -dX);

            if (Math.abs(Math.abs(angle) - Math.PI / 2) < 0.01) {
              leftBusiness++;
              rightBusiness++;
            } else {
              if (Math.abs(angle) < Math.PI / 2) {
                rightBusiness++;
              } else {
                leftBusiness++;
              }
            }
          });

          // Editing the node.
          if (leftBusiness < rightBusiness) {
            node.label = '' + hydrogensLabel + ELEMENTS[node.labelModeData.atomicNumber].symbol;
            node.focusCharacterRange = [hydrogensLabel.length, hydrogensLabel.length + ELEMENTS[node.labelModeData.atomicNumber].symbol.length];
            node.focusLine = 0;
          } else {
            node.label = '' + ELEMENTS[node.labelModeData.atomicNumber].symbol + hydrogensLabel;
            node.focusCharacterRange = [0, ELEMENTS[node.labelModeData.atomicNumber].symbol.length];
            node.focusLine = 0;
          }
        } else {
          // There are no hydrogens.

          //Editing the node.
          node.label = ELEMENTS[node.labelModeData.atomicNumber].symbol;
          node.focusCharacterRange = [0, ELEMENTS[node.labelModeData.atomicNumber].symbol.length];
          node.focusLine = 0;
        }
      }
    }
  });
};

// returns a fragment
var readCdxml = (function (f) {

  // const div = document.createElement('DIV');
  // div.innerHTML = f;


  var parser = new DOMParser();
  var div = parser.parseFromString(f, "application/xml").documentElement;

  var nodes = div.getElementsByTagName('n');

  // f = f.split(/\r?\n/);
  //
  // const countsLine = f[3];
  //
  // const numberOfAtoms = Number(countsLine.substring(0,3));
  // const numberOfBonds = Number(countsLine.substring(3,6));
  //
  var fragment = {
    nodes: Array.prototype.map.call(nodes, function (node) {

      var p = node.getAttribute('p');
      var x = 0,
          y = 0;
      if (p) {
        var pSplit = p.split(' ');
        x = Number(pSplit[0]);
        y = Number(pSplit[1]);
      }

      // const label = node.textContent;

      var atomicNumber = 0;
      if (node.hasAttribute('Element')) {
        var atomicNumberCandidate = Number(node.getAttribute('Element'));
        if (ELEMENTS[atomicNumberCandidate]) {
          atomicNumber = atomicNumberCandidate;
        }
      } else {
        atomicNumber = 6;
      }

      var numberOfHydrogens = Number(node.getAttribute('NumHydrogens'));

      return {
        x: Math.floor(x * 2 + 49999),
        y: Math.floor(y * 2 + 49999),
        z: 49999, //: Math.floor(Number(atomLine.substring(20,30)) * 50),
        label: atomicNumber === 6 ? '' : '' + ELEMENTS[atomicNumber].symbol + (numberOfHydrogens ? 'H' + (numberOfHydrogens === 1 ? '' : numberOfHydrogens) : ''),
        focusCharacterRange: atomicNumber === 6 ? [0, 0] : [0, ELEMENTS[atomicNumber].symbol.length], //[0, label.length], //: [0, atomLine.substring(31,34).trim().length], //[1, 1+atomLine.substring(31,34).trim().length], // one range per line
        focusLine: 0, // assuming one line
        // Floating hydrogens mean that in the -CH2 label, "H2" part occupies
        // the most optimal position (either to the left, above, to the right,
        // or below) so that the least amount of bonds are hindered.
        // labelMode can be:
        // - CHEMICAL_ELEMENT_PLUS_FLOATING_HYDROGENS
        // - MANUAL
        labelMode: 'CHEMICAL_ELEMENT_PLUS_FLOATING_HYDROGENS',
        // If labelMode is MANUAL, then the labelModeData object has no keys
        // and label, focusCharacterRange and focusLine are defined.
        // if labelMode is CHEMICAL_ELEMENT_PLUS_FLOATING_HYDROGENS, the
        // labelModeData has the atomicNumber and numberOfHydrogens keys
        // and based on those, we set and update label, focusCharacterRange and focusLine.
        labelModeData: {
          atomicNumber: atomicNumber, // is a JavaScript number
          numberOfHydrogens: numberOfHydrogens
        },
        beingEdited: false,
        latestMouseDownXY: {
          x: undefined,
          y: undefined
        }
      };
    }),
    bonds: Array.prototype.map.call(div.getElementsByTagName('b'), function (bond) {

      var firstNodeId = bond.getAttribute('B');
      var secondNodeId = bond.getAttribute('E');

      var firstNodeIndex = Array.prototype.findIndex.call(nodes, function (node) {
        return node.getAttribute('id') === firstNodeId;
      });
      var secondNodeIndex = Array.prototype.findIndex.call(nodes, function (node) {
        return node.getAttribute('id') === secondNodeId;
      });

      return {
        firstNodeIndex: firstNodeIndex,
        secondNodeIndex: secondNodeIndex,
        type: 1,
        stereo: 1
      };
    })
  };
  //
  // // reading the atom block
  // for (let i = 4; i < 4 + numberOfAtoms; ++i) {
  // 	const atomLine = f[i];
  //
  //   const node = {
  //     x: Math.floor(Number(atomLine.substring(0,10)) * 50),
  //     y: Math.floor(- Number(atomLine.substring(10,20)) * 50),
  //     z: Math.floor(Number(atomLine.substring(20,30)) * 50),
  //     label: atomLine.substring(31,34).trim()+'H₂₂',
  //     focusCharacterRange: [0, atomLine.substring(31,34).trim().length], //[1, 1+atomLine.substring(31,34).trim().length], // one range per line
  //     focusLine: 0, // assuming one line
  //     beingEdited: false,
  //     latestMouseDownXY: {
  //       x: undefined,
  //       y: undefined
  //     }
  //   };
  //
  //   fragment.nodes.push(node);
  // }
  //
  // // reading the bond block
  // for (let i = 4 + numberOfAtoms; i < 4 + numberOfAtoms + numberOfBonds; ++i) {
  // 	const bondLine = f[i];
  //
  //   const bond = {
  //     firstNodeIndex: Number(bondLine.substring(0,3)) - 1,
  //     secondNodeIndex: Number(bondLine.substring(3,6)) - 1,
  //     type: Number(bondLine.substring(6,9)),
  //     stereo: Number(bondLine.substring(9,12))
  //   };
  //
  //   fragment.bonds.push(bond);
  // }

  positionFloatingHydrogens(fragment);

  return fragment;
});

function getMousePos(canvas, evt) {
  var rect = canvas.getBoundingClientRect();
  // console.log(rect);
  var clientX = void 0,
      clientY = void 0;
  if (evt.changedTouches) {
    // Touch device.
    var touches = evt.changedTouches;
    clientX = touches[0].clientX;
    clientY = touches[0].clientY;
  } else {
    clientX = evt.clientX;
    clientY = evt.clientY;
  }
  return {
    x: Math.ceil(clientX - rect.left),
    y: Math.ceil(clientY - rect.top)
  };
}

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

/**
 * Copyright (c) 2013-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 *
 * 
 */

function makeEmptyFunction(arg) {
  return function () {
    return arg;
  };
}

/**
 * This function accepts and discards inputs; it has no side effects. This is
 * primarily useful idiomatically for overridable function endpoints which
 * always need to be callable, since JS lacks a null-call idiom ala Cocoa.
 */
var emptyFunction = function emptyFunction() {};

emptyFunction.thatReturns = makeEmptyFunction;
emptyFunction.thatReturnsFalse = makeEmptyFunction(false);
emptyFunction.thatReturnsTrue = makeEmptyFunction(true);
emptyFunction.thatReturnsNull = makeEmptyFunction(null);
emptyFunction.thatReturnsThis = function () {
  return this;
};
emptyFunction.thatReturnsArgument = function (arg) {
  return arg;
};

var emptyFunction_1 = emptyFunction;

/**
 * Copyright (c) 2013-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 *
 */

/**
 * Use invariant() to assert state which your program assumes to be true.
 *
 * Provide sprintf-style format (only %s is supported) and arguments
 * to provide information about what broke and what you were
 * expecting.
 *
 * The invariant message will be stripped in production, but the invariant
 * will remain to ensure logic does not differ in production.
 */

var validateFormat = function validateFormat(format) {};

{
  validateFormat = function validateFormat(format) {
    if (format === undefined) {
      throw new Error('invariant requires an error message argument');
    }
  };
}

function invariant(condition, format, a, b, c, d, e, f) {
  validateFormat(format);

  if (!condition) {
    var error;
    if (format === undefined) {
      error = new Error('Minified exception occurred; use the non-minified dev environment ' + 'for the full error message and additional helpful warnings.');
    } else {
      var args = [a, b, c, d, e, f];
      var argIndex = 0;
      error = new Error(format.replace(/%s/g, function () {
        return args[argIndex++];
      }));
      error.name = 'Invariant Violation';
    }

    error.framesToPop = 1; // we don't care about invariant's own frame
    throw error;
  }
}

var invariant_1 = invariant;

/**
 * Similar to invariant but only logs a warning if the condition is not met.
 * This can be used to log issues in development environments in critical
 * paths. Removing the logging code for production environments will keep the
 * same logic and follow the same code paths.
 */

var warning = emptyFunction_1;

{
  var printWarning = function printWarning(format) {
    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    var argIndex = 0;
    var message = 'Warning: ' + format.replace(/%s/g, function () {
      return args[argIndex++];
    });
    if (typeof console !== 'undefined') {
      console.error(message);
    }
    try {
      // --- Welcome to debugging React ---
      // This error was thrown as a convenience so that you can use this stack
      // to find the callsite that caused this warning to fire.
      throw new Error(message);
    } catch (x) {}
  };

  warning = function warning(condition, format) {
    if (format === undefined) {
      throw new Error('`warning(condition, format, ...args)` requires a warning ' + 'message argument');
    }

    if (format.indexOf('Failed Composite propType: ') === 0) {
      return; // Ignore CompositeComponent proptype check.
    }

    if (!condition) {
      for (var _len2 = arguments.length, args = Array(_len2 > 2 ? _len2 - 2 : 0), _key2 = 2; _key2 < _len2; _key2++) {
        args[_key2 - 2] = arguments[_key2];
      }

      printWarning.apply(undefined, [format].concat(args));
    }
  };
}

var warning_1 = warning;

/**
 * Copyright 2013-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */

var ReactPropTypesSecret = 'SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED';

var ReactPropTypesSecret_1 = ReactPropTypesSecret;

{
  var invariant$1 = invariant_1;
  var warning$1 = warning_1;
  var ReactPropTypesSecret$1 = ReactPropTypesSecret_1;
  var loggedTypeFailures = {};
}

/**
 * Assert that the values match with the type specs.
 * Error messages are memorized and will only be shown once.
 *
 * @param {object} typeSpecs Map of name to a ReactPropType
 * @param {object} values Runtime values that need to be type-checked
 * @param {string} location e.g. "prop", "context", "child context"
 * @param {string} componentName Name of the component for error messages.
 * @param {?Function} getStack Returns the component stack.
 * @private
 */
function checkPropTypes(typeSpecs, values, location, componentName, getStack) {
  {
    for (var typeSpecName in typeSpecs) {
      if (typeSpecs.hasOwnProperty(typeSpecName)) {
        var error;
        // Prop type validation may throw. In case they do, we don't want to
        // fail the render phase where it didn't fail before. So we log it.
        // After these have been cleaned up, we'll let them throw.
        try {
          // This is intentionally an invariant that gets caught. It's the same
          // behavior as without this statement except with a better message.
          invariant$1(typeof typeSpecs[typeSpecName] === 'function', '%s: %s type `%s` is invalid; it must be a function, usually from ' + 'React.PropTypes.', componentName || 'React class', location, typeSpecName);
          error = typeSpecs[typeSpecName](values, typeSpecName, componentName, location, null, ReactPropTypesSecret$1);
        } catch (ex) {
          error = ex;
        }
        warning$1(!error || error instanceof Error, '%s: type specification of %s `%s` is invalid; the type checker ' + 'function must return `null` or an `Error` but returned a %s. ' + 'You may have forgotten to pass an argument to the type checker ' + 'creator (arrayOf, instanceOf, objectOf, oneOf, oneOfType, and ' + 'shape all require an argument).', componentName || 'React class', location, typeSpecName, typeof error);
        if (error instanceof Error && !(error.message in loggedTypeFailures)) {
          // Only monitor this failure once because there tends to be a lot of the
          // same error.
          loggedTypeFailures[error.message] = true;

          var stack = getStack ? getStack() : '';

          warning$1(false, 'Failed %s type: %s%s', location, error.message, stack != null ? stack : '');
        }
      }
    }
  }
}

var checkPropTypes_1 = checkPropTypes;

var factoryWithTypeCheckers = function(isValidElement, throwOnDirectAccess) {
  /* global Symbol */
  var ITERATOR_SYMBOL = typeof Symbol === 'function' && Symbol.iterator;
  var FAUX_ITERATOR_SYMBOL = '@@iterator'; // Before Symbol spec.

  /**
   * Returns the iterator method function contained on the iterable object.
   *
   * Be sure to invoke the function with the iterable as context:
   *
   *     var iteratorFn = getIteratorFn(myIterable);
   *     if (iteratorFn) {
   *       var iterator = iteratorFn.call(myIterable);
   *       ...
   *     }
   *
   * @param {?object} maybeIterable
   * @return {?function}
   */
  function getIteratorFn(maybeIterable) {
    var iteratorFn = maybeIterable && (ITERATOR_SYMBOL && maybeIterable[ITERATOR_SYMBOL] || maybeIterable[FAUX_ITERATOR_SYMBOL]);
    if (typeof iteratorFn === 'function') {
      return iteratorFn;
    }
  }

  /**
   * Collection of methods that allow declaration and validation of props that are
   * supplied to React components. Example usage:
   *
   *   var Props = require('ReactPropTypes');
   *   var MyArticle = React.createClass({
   *     propTypes: {
   *       // An optional string prop named "description".
   *       description: Props.string,
   *
   *       // A required enum prop named "category".
   *       category: Props.oneOf(['News','Photos']).isRequired,
   *
   *       // A prop named "dialog" that requires an instance of Dialog.
   *       dialog: Props.instanceOf(Dialog).isRequired
   *     },
   *     render: function() { ... }
   *   });
   *
   * A more formal specification of how these methods are used:
   *
   *   type := array|bool|func|object|number|string|oneOf([...])|instanceOf(...)
   *   decl := ReactPropTypes.{type}(.isRequired)?
   *
   * Each and every declaration produces a function with the same signature. This
   * allows the creation of custom validation functions. For example:
   *
   *  var MyLink = React.createClass({
   *    propTypes: {
   *      // An optional string or URI prop named "href".
   *      href: function(props, propName, componentName) {
   *        var propValue = props[propName];
   *        if (propValue != null && typeof propValue !== 'string' &&
   *            !(propValue instanceof URI)) {
   *          return new Error(
   *            'Expected a string or an URI for ' + propName + ' in ' +
   *            componentName
   *          );
   *        }
   *      }
   *    },
   *    render: function() {...}
   *  });
   *
   * @internal
   */

  var ANONYMOUS = '<<anonymous>>';

  // Important!
  // Keep this list in sync with production version in `./factoryWithThrowingShims.js`.
  var ReactPropTypes = {
    array: createPrimitiveTypeChecker('array'),
    bool: createPrimitiveTypeChecker('boolean'),
    func: createPrimitiveTypeChecker('function'),
    number: createPrimitiveTypeChecker('number'),
    object: createPrimitiveTypeChecker('object'),
    string: createPrimitiveTypeChecker('string'),
    symbol: createPrimitiveTypeChecker('symbol'),

    any: createAnyTypeChecker(),
    arrayOf: createArrayOfTypeChecker,
    element: createElementTypeChecker(),
    instanceOf: createInstanceTypeChecker,
    node: createNodeChecker(),
    objectOf: createObjectOfTypeChecker,
    oneOf: createEnumTypeChecker,
    oneOfType: createUnionTypeChecker,
    shape: createShapeTypeChecker
  };

  /**
   * inlined Object.is polyfill to avoid requiring consumers ship their own
   * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/is
   */
  /*eslint-disable no-self-compare*/
  function is(x, y) {
    // SameValue algorithm
    if (x === y) {
      // Steps 1-5, 7-10
      // Steps 6.b-6.e: +0 != -0
      return x !== 0 || 1 / x === 1 / y;
    } else {
      // Step 6.a: NaN == NaN
      return x !== x && y !== y;
    }
  }
  /*eslint-enable no-self-compare*/

  /**
   * We use an Error-like object for backward compatibility as people may call
   * PropTypes directly and inspect their output. However, we don't use real
   * Errors anymore. We don't inspect their stack anyway, and creating them
   * is prohibitively expensive if they are created too often, such as what
   * happens in oneOfType() for any type before the one that matched.
   */
  function PropTypeError(message) {
    this.message = message;
    this.stack = '';
  }
  // Make `instanceof Error` still work for returned errors.
  PropTypeError.prototype = Error.prototype;

  function createChainableTypeChecker(validate) {
    {
      var manualPropTypeCallCache = {};
      var manualPropTypeWarningCount = 0;
    }
    function checkType(isRequired, props, propName, componentName, location, propFullName, secret) {
      componentName = componentName || ANONYMOUS;
      propFullName = propFullName || propName;

      if (secret !== ReactPropTypesSecret_1) {
        if (throwOnDirectAccess) {
          // New behavior only for users of `prop-types` package
          invariant_1(
            false,
            'Calling PropTypes validators directly is not supported by the `prop-types` package. ' +
            'Use `PropTypes.checkPropTypes()` to call them. ' +
            'Read more at http://fb.me/use-check-prop-types'
          );
        } else if ("development" !== 'production' && typeof console !== 'undefined') {
          // Old behavior for people using React.PropTypes
          var cacheKey = componentName + ':' + propName;
          if (
            !manualPropTypeCallCache[cacheKey] &&
            // Avoid spamming the console because they are often not actionable except for lib authors
            manualPropTypeWarningCount < 3
          ) {
            warning_1(
              false,
              'You are manually calling a React.PropTypes validation ' +
              'function for the `%s` prop on `%s`. This is deprecated ' +
              'and will throw in the standalone `prop-types` package. ' +
              'You may be seeing this warning due to a third-party PropTypes ' +
              'library. See https://fb.me/react-warning-dont-call-proptypes ' + 'for details.',
              propFullName,
              componentName
            );
            manualPropTypeCallCache[cacheKey] = true;
            manualPropTypeWarningCount++;
          }
        }
      }
      if (props[propName] == null) {
        if (isRequired) {
          if (props[propName] === null) {
            return new PropTypeError('The ' + location + ' `' + propFullName + '` is marked as required ' + ('in `' + componentName + '`, but its value is `null`.'));
          }
          return new PropTypeError('The ' + location + ' `' + propFullName + '` is marked as required in ' + ('`' + componentName + '`, but its value is `undefined`.'));
        }
        return null;
      } else {
        return validate(props, propName, componentName, location, propFullName);
      }
    }

    var chainedCheckType = checkType.bind(null, false);
    chainedCheckType.isRequired = checkType.bind(null, true);

    return chainedCheckType;
  }

  function createPrimitiveTypeChecker(expectedType) {
    function validate(props, propName, componentName, location, propFullName, secret) {
      var propValue = props[propName];
      var propType = getPropType(propValue);
      if (propType !== expectedType) {
        // `propValue` being instance of, say, date/regexp, pass the 'object'
        // check, but we can offer a more precise error message here rather than
        // 'of type `object`'.
        var preciseType = getPreciseType(propValue);

        return new PropTypeError('Invalid ' + location + ' `' + propFullName + '` of type ' + ('`' + preciseType + '` supplied to `' + componentName + '`, expected ') + ('`' + expectedType + '`.'));
      }
      return null;
    }
    return createChainableTypeChecker(validate);
  }

  function createAnyTypeChecker() {
    return createChainableTypeChecker(emptyFunction_1.thatReturnsNull);
  }

  function createArrayOfTypeChecker(typeChecker) {
    function validate(props, propName, componentName, location, propFullName) {
      if (typeof typeChecker !== 'function') {
        return new PropTypeError('Property `' + propFullName + '` of component `' + componentName + '` has invalid PropType notation inside arrayOf.');
      }
      var propValue = props[propName];
      if (!Array.isArray(propValue)) {
        var propType = getPropType(propValue);
        return new PropTypeError('Invalid ' + location + ' `' + propFullName + '` of type ' + ('`' + propType + '` supplied to `' + componentName + '`, expected an array.'));
      }
      for (var i = 0; i < propValue.length; i++) {
        var error = typeChecker(propValue, i, componentName, location, propFullName + '[' + i + ']', ReactPropTypesSecret_1);
        if (error instanceof Error) {
          return error;
        }
      }
      return null;
    }
    return createChainableTypeChecker(validate);
  }

  function createElementTypeChecker() {
    function validate(props, propName, componentName, location, propFullName) {
      var propValue = props[propName];
      if (!isValidElement(propValue)) {
        var propType = getPropType(propValue);
        return new PropTypeError('Invalid ' + location + ' `' + propFullName + '` of type ' + ('`' + propType + '` supplied to `' + componentName + '`, expected a single ReactElement.'));
      }
      return null;
    }
    return createChainableTypeChecker(validate);
  }

  function createInstanceTypeChecker(expectedClass) {
    function validate(props, propName, componentName, location, propFullName) {
      if (!(props[propName] instanceof expectedClass)) {
        var expectedClassName = expectedClass.name || ANONYMOUS;
        var actualClassName = getClassName(props[propName]);
        return new PropTypeError('Invalid ' + location + ' `' + propFullName + '` of type ' + ('`' + actualClassName + '` supplied to `' + componentName + '`, expected ') + ('instance of `' + expectedClassName + '`.'));
      }
      return null;
    }
    return createChainableTypeChecker(validate);
  }

  function createEnumTypeChecker(expectedValues) {
    if (!Array.isArray(expectedValues)) {
      warning_1(false, 'Invalid argument supplied to oneOf, expected an instance of array.');
      return emptyFunction_1.thatReturnsNull;
    }

    function validate(props, propName, componentName, location, propFullName) {
      var propValue = props[propName];
      for (var i = 0; i < expectedValues.length; i++) {
        if (is(propValue, expectedValues[i])) {
          return null;
        }
      }

      var valuesString = JSON.stringify(expectedValues);
      return new PropTypeError('Invalid ' + location + ' `' + propFullName + '` of value `' + propValue + '` ' + ('supplied to `' + componentName + '`, expected one of ' + valuesString + '.'));
    }
    return createChainableTypeChecker(validate);
  }

  function createObjectOfTypeChecker(typeChecker) {
    function validate(props, propName, componentName, location, propFullName) {
      if (typeof typeChecker !== 'function') {
        return new PropTypeError('Property `' + propFullName + '` of component `' + componentName + '` has invalid PropType notation inside objectOf.');
      }
      var propValue = props[propName];
      var propType = getPropType(propValue);
      if (propType !== 'object') {
        return new PropTypeError('Invalid ' + location + ' `' + propFullName + '` of type ' + ('`' + propType + '` supplied to `' + componentName + '`, expected an object.'));
      }
      for (var key in propValue) {
        if (propValue.hasOwnProperty(key)) {
          var error = typeChecker(propValue, key, componentName, location, propFullName + '.' + key, ReactPropTypesSecret_1);
          if (error instanceof Error) {
            return error;
          }
        }
      }
      return null;
    }
    return createChainableTypeChecker(validate);
  }

  function createUnionTypeChecker(arrayOfTypeCheckers) {
    if (!Array.isArray(arrayOfTypeCheckers)) {
      warning_1(false, 'Invalid argument supplied to oneOfType, expected an instance of array.');
      return emptyFunction_1.thatReturnsNull;
    }

    for (var i = 0; i < arrayOfTypeCheckers.length; i++) {
      var checker = arrayOfTypeCheckers[i];
      if (typeof checker !== 'function') {
        warning_1(
          false,
          'Invalid argument supplid to oneOfType. Expected an array of check functions, but ' +
          'received %s at index %s.',
          getPostfixForTypeWarning(checker),
          i
        );
        return emptyFunction_1.thatReturnsNull;
      }
    }

    function validate(props, propName, componentName, location, propFullName) {
      for (var i = 0; i < arrayOfTypeCheckers.length; i++) {
        var checker = arrayOfTypeCheckers[i];
        if (checker(props, propName, componentName, location, propFullName, ReactPropTypesSecret_1) == null) {
          return null;
        }
      }

      return new PropTypeError('Invalid ' + location + ' `' + propFullName + '` supplied to ' + ('`' + componentName + '`.'));
    }
    return createChainableTypeChecker(validate);
  }

  function createNodeChecker() {
    function validate(props, propName, componentName, location, propFullName) {
      if (!isNode(props[propName])) {
        return new PropTypeError('Invalid ' + location + ' `' + propFullName + '` supplied to ' + ('`' + componentName + '`, expected a ReactNode.'));
      }
      return null;
    }
    return createChainableTypeChecker(validate);
  }

  function createShapeTypeChecker(shapeTypes) {
    function validate(props, propName, componentName, location, propFullName) {
      var propValue = props[propName];
      var propType = getPropType(propValue);
      if (propType !== 'object') {
        return new PropTypeError('Invalid ' + location + ' `' + propFullName + '` of type `' + propType + '` ' + ('supplied to `' + componentName + '`, expected `object`.'));
      }
      for (var key in shapeTypes) {
        var checker = shapeTypes[key];
        if (!checker) {
          continue;
        }
        var error = checker(propValue, key, componentName, location, propFullName + '.' + key, ReactPropTypesSecret_1);
        if (error) {
          return error;
        }
      }
      return null;
    }
    return createChainableTypeChecker(validate);
  }

  function isNode(propValue) {
    switch (typeof propValue) {
      case 'number':
      case 'string':
      case 'undefined':
        return true;
      case 'boolean':
        return !propValue;
      case 'object':
        if (Array.isArray(propValue)) {
          return propValue.every(isNode);
        }
        if (propValue === null || isValidElement(propValue)) {
          return true;
        }

        var iteratorFn = getIteratorFn(propValue);
        if (iteratorFn) {
          var iterator = iteratorFn.call(propValue);
          var step;
          if (iteratorFn !== propValue.entries) {
            while (!(step = iterator.next()).done) {
              if (!isNode(step.value)) {
                return false;
              }
            }
          } else {
            // Iterator will provide entry [k,v] tuples rather than values.
            while (!(step = iterator.next()).done) {
              var entry = step.value;
              if (entry) {
                if (!isNode(entry[1])) {
                  return false;
                }
              }
            }
          }
        } else {
          return false;
        }

        return true;
      default:
        return false;
    }
  }

  function isSymbol(propType, propValue) {
    // Native Symbol.
    if (propType === 'symbol') {
      return true;
    }

    // 19.4.3.5 Symbol.prototype[@@toStringTag] === 'Symbol'
    if (propValue['@@toStringTag'] === 'Symbol') {
      return true;
    }

    // Fallback for non-spec compliant Symbols which are polyfilled.
    if (typeof Symbol === 'function' && propValue instanceof Symbol) {
      return true;
    }

    return false;
  }

  // Equivalent of `typeof` but with special handling for array and regexp.
  function getPropType(propValue) {
    var propType = typeof propValue;
    if (Array.isArray(propValue)) {
      return 'array';
    }
    if (propValue instanceof RegExp) {
      // Old webkits (at least until Android 4.0) return 'function' rather than
      // 'object' for typeof a RegExp. We'll normalize this here so that /bla/
      // passes PropTypes.object.
      return 'object';
    }
    if (isSymbol(propType, propValue)) {
      return 'symbol';
    }
    return propType;
  }

  // This handles more types than `getPropType`. Only used for error messages.
  // See `createPrimitiveTypeChecker`.
  function getPreciseType(propValue) {
    if (typeof propValue === 'undefined' || propValue === null) {
      return '' + propValue;
    }
    var propType = getPropType(propValue);
    if (propType === 'object') {
      if (propValue instanceof Date) {
        return 'date';
      } else if (propValue instanceof RegExp) {
        return 'regexp';
      }
    }
    return propType;
  }

  // Returns a string that is postfixed to a warning about an invalid type.
  // For example, "undefined" or "of type array"
  function getPostfixForTypeWarning(value) {
    var type = getPreciseType(value);
    switch (type) {
      case 'array':
      case 'object':
        return 'an ' + type;
      case 'boolean':
      case 'date':
      case 'regexp':
        return 'a ' + type;
      default:
        return type;
    }
  }

  // Returns class name of the object, if any.
  function getClassName(propValue) {
    if (!propValue.constructor || !propValue.constructor.name) {
      return ANONYMOUS;
    }
    return propValue.constructor.name;
  }

  ReactPropTypes.checkPropTypes = checkPropTypes_1;
  ReactPropTypes.PropTypes = ReactPropTypes;

  return ReactPropTypes;
};

var index = createCommonjsModule(function (module) {
/**
 * Copyright 2013-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */

{
  var REACT_ELEMENT_TYPE = (typeof Symbol === 'function' &&
    Symbol.for &&
    Symbol.for('react.element')) ||
    0xeac7;

  var isValidElement = function(object) {
    return typeof object === 'object' &&
      object !== null &&
      object.$$typeof === REACT_ELEMENT_TYPE;
  };

  // By explicitly using `prop-types` you are opting into new development behavior.
  // http://fb.me/prop-types-in-prod
  var throwOnDirectAccess = true;
  module.exports = factoryWithTypeCheckers(isValidElement, throwOnDirectAccess);
}
});

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};









var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};



var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};











var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

var Node = function (_Component) {
  inherits(Node, _Component);

  function Node() {
    classCallCheck(this, Node);
    return possibleConstructorReturn(this, _Component.apply(this, arguments));
  }

  Node.prototype.render = function render() {
    var c = getCanvas(this.props.node, this.props.canvasCache);
    if (c) {
      // const width = Math.max(...c.filledPixels.map(xy => xy[0])) - Math.min(...c.filledPixels.map(xy => xy[0]));
      // const height = Math.max(...c.filledPixels.map(xy => xy[1])) - Math.min(...c.filledPixels.map(xy => xy[1]));
      return React__default.createElement(
        'g',
        { transform: 'translate(' + this.props.node.x + ' ' + this.props.node.y + ')' },
        React__default.createElement(
          'text',
          {
            style: {
              fill: 'black',
              font: CONSTANTS.FONT_SIZE + 'px sans-serif',
              WebkitTouchCallout: 'none',
              WebkitUserSelect: 'none',
              KhtmlUserSelect: 'none',
              MozUserSelect: 'none',
              MsUserSelect: 'none',
              userSelect: 'none'
            },
            x: c.labelOriginXRelativeToFocus,
            y: c.labelOriginYRelativeToFocus
          },
          this.props.node.label
        )
      );
    }

    return React__default.createElement('rect', {
      x: this.props.node.x,
      y: this.props.node.y,
      width: 6,
      height: 6,
      style: {
        fill: 'red'
      }
    });
  };

  return Node;
}(React.Component);

Node.propTypes = {
  canvasCache: index.array.isRequired,
  node: index.object.isRequired
};

function getSingleBondPath(bond, firstNode, secondNode) {
  var d = subtractPoints(secondNode, firstNode);
  var length = Math.sqrt(d.x * d.x + d.y * d.y);
  var bondVerticalUnitVector = createPoint(0 - d.y / length, d.x / length);

  var firstEndPath = void 0;
  if (firstNode.label) {
    firstEndPath = pointToPathString(slidePoint(addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2)), addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2)), bond.firstT)) + ' L ' + pointToPathString(slidePoint(addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2)), addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2)), bond.firstT));
  } else {
    firstEndPath = pointToPathString(slidePoint(addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2)), addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2)), bond.firstUpT)) + ' L ' + pointToPathString(firstNode) + ' L ' + pointToPathString(slidePoint(addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2)), addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2)), bond.firstDownT));
  }

  var secondEndPath = void 0;
  if (secondNode.label) {
    secondEndPath = pointToPathString(slidePoint(addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2)), addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2)), bond.secondT)) + ' L ' + pointToPathString(slidePoint(addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2)), addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2)), bond.secondT));
  } else {
    secondEndPath = pointToPathString(slidePoint(addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2)), addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH / 2)), bond.secondDownT)) + ' L ' + pointToPathString(secondNode) + ' L ' + pointToPathString(slidePoint(addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2)), addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH / 2)), bond.secondUpT));
  }

  return 'M ' + firstEndPath + ' L ' + secondEndPath + ' Z';
}

var Bond = function (_Component) {
  inherits(Bond, _Component);

  function Bond() {
    classCallCheck(this, Bond);
    return possibleConstructorReturn(this, _Component.apply(this, arguments));
  }

  Bond.prototype.render = function render() {

    var pathD = getSingleBondPath(this.props.bond, this.props.firstNode, this.props.secondNode);

    return React__default.createElement('path', {
      d: pathD,
      style: {
        fill: 'black'
      }
    });
  };

  return Bond;
}(React.Component);

Bond.propTypes = {
  bond: index.object.isRequired,
  firstNode: index.object.isRequired,
  secondNode: index.object.isRequired
};

var Fragment = function (_Component) {
  inherits(Fragment, _Component);

  function Fragment() {
    classCallCheck(this, Fragment);
    return possibleConstructorReturn(this, _Component.apply(this, arguments));
  }

  Fragment.prototype.render = function render() {
    var _this2 = this;

    return React__default.createElement(
      'g',
      null,
      this.props.fragment.nodes.map(function (node, nodeIndex) {
        return React__default.createElement(Node, {
          key: nodeIndex,
          canvasCache: _this2.props.canvasCache,
          node: node
        });
      }),
      this.props.fragment.bonds.map(function (bond, bondIndex) {
        return React__default.createElement(Bond, {
          key: bondIndex,
          bond: bond,
          firstNode: _this2.props.fragment.nodes[bond.firstNodeIndex],
          secondNode: _this2.props.fragment.nodes[bond.secondNodeIndex]
        });
      })
    );
  };

  return Fragment;
}(React.Component);

Fragment.propTypes = {
  canvasCache: index.array.isRequired,
  fragment: index.object.isRequired
};

// We use contentEditable span and not input.
// Benefits of input:
// (a) no formatting,
// (b) always single line,
// (c) easier onChange handlers?
// Benefits of contentEditable:
// (a) stretches on demand,
// (b) we can house formatted content, i.e. it's future proofed

// Need to copy the way draft js handles onpaste (we don't want to paste
// formatted content), entities (our focus range is an entity)
// Use mutation observer (instead of onInput)?

var Label = function (_Component) {
  inherits(Label, _Component);

  function Label() {
    classCallCheck(this, Label);
    return possibleConstructorReturn(this, _Component.apply(this, arguments));
  }

  Label.prototype.componentDidMount = function componentDidMount() {
    this.refs.span.textContent = this.props.value;
  };

  // componentDidUpdate(prevProps, prevState) {
  //
  //   const div = document.createElement('DIV');
  //   div.textContent = this.props.value;
  //
  //   if (this.refs.span.innerHTML !== div.innerHTML) {
  //     this.refs.span.innerHTML = div.innerHTML;
  //   }
  // }


  Label.prototype.render = function render() {
    var _this2 = this;

    return React__default.createElement(
      'div',
      { style: {
          position: 'absolute',
          left: this.props.origin.x + 'px',
          top: this.props.origin.y - 5 * CONSTANTS.FONT_SIZE + 'px'
        } },
      React__default.createElement('span', {
        ref: 'span',
        style: {
          display: 'inline-block',
          padding: '10px 0',

          fontSize: CONSTANTS.FONT_SIZE + 'px',

          opacity: '0.75',

          border: 'none',
          outline: 'none',
          appearance: 'none',
          WebkitAppearance: 'none',
          backgroundColor: 'transparent',
          font: CONSTANTS.FONT_SIZE + 'px sans-serif',

          lineHeight: CONSTANTS.FONT_SIZE + 'px',
          whiteSpace: 'nowrap'
        },
        onInput: function onInput() {
          _this2.props.onChange(_this2.refs.span.textContent);
        }
        // onKeyDown={() => {
        //   setTimeout(() => {
        //     this.props.onChange(this.refs.span.textContent);
        //   }, 1);
        // }}
        , contentEditable: true
      }),
      React__default.createElement('span', { style: {
          display: 'inline-block',
          height: 5 * CONSTANTS.FONT_SIZE + 'px'
        } })
    );
  };

  return Label;
}(React.Component);

Label.propTypes = {
  origin: index.object.isRequired, // intersection of left edge and baseline
  onChange: index.func.isRequired, // accepts one parameter which is new label
  value: index.string.isRequired
};

// The function returns an array of canvases that don't exist in canvasCache.
var create2MissingCanvases = function (nodes, canvasCache) {
  var newCs = [];

  nodes.forEach(function (node) {

    var existingC = getCanvas(node, [].concat(canvasCache, newCs));
    if (existingC) return;

    var canvasDomElement = document.createElement('CANVAS');
    var ctx = canvasDomElement.getContext('2d');
    ctx.font = CONSTANTS.FONT_SIZE + 'px sans-serif';
    ctx.strokeStyle = 'red';
    ctx.lineWidth = CONSTANTS.LABEL_BOND_SPACING;
    ctx.lineJoin = 'round';
    // const nodeLabelLines = node.label.split('\n');

    // Assuming one line. (We call it label for now.)
    // Assuming no kerning or ligatures.
    // Y is directed downwards.
    // We will have one focus character range per line; center of the range is the center of the line.
    // Origin is the point we specify when using ctx.fillText/strokeText (left bottom point of text).
    // Bottom of the label is the baseline. Height is font size.
    var labelOriginXRelativeToFocus = -ctx.measureText(node.label.substring(0, node.focusCharacterRange[0])).width - ctx.measureText(node.label.substring(node.focusCharacterRange[0], node.focusCharacterRange[1])).width / 2;
    var labelOriginYRelativeToFocus = CONSTANTS.FONT_SIZE / 3;
    var labelWidth = ctx.measureText(node.label).width;
    var labelHeight = CONSTANTS.FONT_SIZE;

    // Adding padding to our canvases for "ygp..." protrusions etc.
    var leftCanvasBorderXRelativeToFocus = labelOriginXRelativeToFocus - CONSTANTS.ADDED_MARGIN_FOR_CANVASES;
    var rightCanvasBorderXRelativeToFocus = labelOriginXRelativeToFocus + labelWidth + CONSTANTS.ADDED_MARGIN_FOR_CANVASES;
    var topCanvasBorderYRelativeToFocus = labelOriginYRelativeToFocus - labelHeight - CONSTANTS.ADDED_MARGIN_FOR_CANVASES;
    var bottomCanvasBorderYRelativeToFocus = labelOriginYRelativeToFocus + CONSTANTS.ADDED_MARGIN_FOR_CANVASES;

    // const textWidth = Math.max(...nodeLabelLines.map(labelLine => ctx.measureText(labelLine).width));
    // console.log(textWidth);
    // const textHeight = nodeLabelLines.length * CONSTANTS.FONT_SIZE;
    canvasDomElement.width = rightCanvasBorderXRelativeToFocus - leftCanvasBorderXRelativeToFocus;
    canvasDomElement.height = bottomCanvasBorderYRelativeToFocus - topCanvasBorderYRelativeToFocus;

    // Our ctx got reset after we resized the canvas.
    ctx.font = CONSTANTS.FONT_SIZE + 'px sans-serif';
    ctx.strokeStyle = 'red';
    ctx.lineWidth = CONSTANTS.LABEL_BOND_SPACING;
    ctx.lineJoin = 'round';

    // nodeLabelLines.forEach((labelLine, labelLineIndex) => {
    //   ctx.strokeText(labelLine, ADDED_MARGIN_FOR_CANVASES, ADDED_MARGIN_FOR_CANVASES + (labelLineIndex + 1) * CONSTANTS.FONT_SIZE);
    // });
    ctx.strokeText(node.label, labelOriginXRelativeToFocus - leftCanvasBorderXRelativeToFocus, labelOriginYRelativeToFocus - topCanvasBorderYRelativeToFocus);

    var pixelData = ctx.getImageData(0, 0, canvasDomElement.width, canvasDomElement.height).data;

    var filledPixels = [];
    for (var x = 0; x < canvasDomElement.width; ++x) {
      for (var y = 0; y < canvasDomElement.height; ++y) {
        if (pixelData[y * canvasDomElement.width * 4 + x * 4 + 3] > 0) {
          // max for alpha is 255?
          filledPixels.push([x + leftCanvasBorderXRelativeToFocus + 0.5, y + topCanvasBorderYRelativeToFocus + 0.5]);
        }
      }
    }

    // The Y coordinate line is directed downwards.
    var minX = 99999; // We will apply Math.min to it.
    var minY = 99999; // We will apply Math.min to it.
    var maxX = 0; // We will apply Math.max to it.
    var maxY = 0; // // We will apply Math.max to it.

    filledPixels.forEach(function (filledPixel) {
      minX = Math.min(minX, filledPixel[0]);
      minY = Math.min(minY, filledPixel[1]);
      maxX = Math.max(maxX, filledPixel[0]);
      maxY = Math.max(maxY, filledPixel[1]);
    });

    var filledPixelsFlat = filledPixels.reduce(function (acc, filledPixel) {
      acc.push(filledPixel[0]);
      acc.push(filledPixel[1]);
      return acc;
    }, []);
    var filledPixelsData = new Float32Array(filledPixelsFlat);

    var c = {
      label: node.label,
      focusCharacterRange: node.focusCharacterRange, // one range per line
      focusLine: node.focusLine,
      filledPixels: filledPixels,
      filledPixelsData: filledPixelsData, // origin here is the focus point which is in the middle of focus line (both in X and in Y coordinates)
      labelOriginXRelativeToFocus: labelOriginXRelativeToFocus,
      labelOriginYRelativeToFocus: labelOriginYRelativeToFocus,
      minX: minX,
      minY: minY,
      maxX: maxX,
      maxY: maxY
    };

    // this.setState(state => {
    //   // if (state.canvasCache.some(c => {
    //   //   return c.label === node.label;
    //   // })) {
    //   //   return state;
    //   // }
    //   return {
    //     ...state,
    //     canvasCache: [...state.canvasCache, c]
    //   };
    // });


    newCs.push(c);
    //return c;
  });
  return newCs;
};

// An object can be anything: a node, a bond, or even an entire fragment!

// An object path goes like this:
// ['fragments', 3, 'nodes', 5]


function _pointerDownHandler(event) {

  var mousePos = getMousePos(this.refs.svg, event);
  var shiftKeyIsBeingPressed = event.shiftKey;

  this.setState(function (state) {

    mousePos.x += state.minX;
    mousePos.y += state.minY;

    var stateAssigns = {
      latestMouseDownPoint: mousePos,
      mouseIsDown: true,
      objects: _extends({}, state.objects, {
        fragments: state.objects.fragments.map(function (fragment) {
          return _extends({}, fragment, {
            nodes: fragment.nodes.map(function (node) {
              return _extends({}, node, {
                latestMouseDownXY: {
                  x: node.x,
                  y: node.y
                }
              });
            })
          });
        })
      })
    };

    if (shiftKeyIsBeingPressed) {
      var pathOfTheObjectBeingPointedAt = getPathOfTheObjectBeingPointedAt(mousePos, state.objects);

      if (pathOfTheObjectBeingPointedAt) {

        if (arrayOfArraysIncludesArray(state.pathsOfSelectedObjects, pathOfTheObjectBeingPointedAt)) {
          // We are hitting one of the selected objects
          return _extends({}, state, stateAssigns);
        }

        // We are hitting something unselected.
        // What we are hitting is selected in addition to the objects that have already been slected (if any)
        // because the Shift key is being pressed.
        return _extends({}, state, {
          pathsOfSelectedObjects: [].concat(state.pathsOfSelectedObjects, [pathOfTheObjectBeingPointedAt])
        }, stateAssigns);
      }
    } else {
      var _pathOfTheObjectBeingPointedAt = getPathOfTheObjectBeingPointedAt(mousePos, state.objects);

      if (_pathOfTheObjectBeingPointedAt) {

        if (arrayOfArraysIncludesArray(state.pathsOfSelectedObjects, _pathOfTheObjectBeingPointedAt)) {
          // We are hitting one of the selected objects
          return _extends({}, state, stateAssigns);
        }

        // We are hitting something unselected.
        // If any objects are selected, they are deselected
        // because the Shift key is not pressed.
        // What we are hitting is selected.
        return _extends({}, state, {
          pathsOfSelectedObjects: [_pathOfTheObjectBeingPointedAt]
        }, stateAssigns);
      }

      // We are hitting nowhere.
      return _extends({}, state, {
        pathsOfSelectedObjects: []
      }, stateAssigns);
    }

    return _extends({}, state, stateAssigns);
  });
}

function _pointerUpHandler(event) {

  var mousePos = getMousePos(this.refs.svg, event);

  this.setState(function (state) {

    mousePos.x += state.minX;
    mousePos.y += state.minY;

    // let hitFragmentIndex = undefined;
    // let hitNodeIndex = undefined;
    //
    // state.objects.fragments.forEach((fragment, fragmentIndex) => {
    //   fragment.nodes.forEach((node, nodeIndex) => {
    //     if (Math.abs(node.x - mousePos.x) < 20 && Math.abs(node.y - mousePos.y) < 20) {
    //       hitFragmentIndex = fragmentIndex;
    //       hitNodeIndex = nodeIndex;
    //     }
    //   });
    // });


    if (state.latestMouseDownPoint.x === mousePos.x && state.latestMouseDownPoint.y === mousePos.y) {}
    // We clicked and released the pointer button.

    // const pathOfTheObjectHavingBeenClicked = getPathOfTheObjectBeingPointedAt(mousePos, state.objects);
    //
    // if (pathOfTheObjectHavingBeenClicked) {
    //   return {
    //     ...state,
    //     pathsOfSelectedObjects: [
    //       pathOfTheObjectHavingBeenClicked
    //     ],
    //     mouseIsDown: false
    //   };
    // }
    //
    // // We clicked nowhere.
    // return {
    //   ...state,
    //   pathsOfSelectedObjects: [],
    //   mouseIsDown: false
    // };


    // return {
    //   ...state,
    //   objects: {
    //     ...state.objects,
    //     fragments: state.objects.fragments.map((fragment, fragmentIndex) => {
    //       return {
    //         ...fragment,
    //         nodes: fragment.nodes.map((node, nodeIndex) => {
    //           return {
    //             ...node,
    //             beingEdited: (fragmentIndex === hitFragmentIndex && nodeIndex === hitNodeIndex)
    //           };
    //         })
    //       };
    //     })
    //   }
    // };

    // if (typeof hitFragmentIndex !== 'undefined' && typeof hitNodeIndex !== 'undefined') {
    //   // we clicked on a node
    // }

    // No changes.
    return _extends({}, state, {
      mouseIsDown: false
    });
  });
}

function _pointerMoveHandler(event) {

  var state = this.state;
  if (state.mouseIsDown && state.pathsOfSelectedObjects.length) {
    // Mouse is pressed and something is selected, so we are dragging
    // the selected objects.
    event.preventDefault();
  }

  var mousePos = getMousePos(this.refs.svg, event);
  this.setState(function (state) {

    mousePos.x += state.minX;
    mousePos.y += state.minY;

    if (state.mouseIsDown) {
      if (state.pathsOfSelectedObjects.length) {
        // Mouse is pressed and something is selected, so we are dragging
        // the selected objects.

        var newCanvasCache = [].concat(state.canvasCache);

        var newObjects = _extends({}, state.objects, {
          fragments: state.objects.fragments.map(function (fragment, fragmentIndex) {
            var newFragment = _extends({}, fragment, {
              nodes: fragment.nodes.map(function (node, nodeIndex) {
                if (arrayOfArraysIncludesArray(state.pathsOfSelectedObjects, ['fragments', fragmentIndex, 'nodes', nodeIndex])) {
                  return _extends({}, node, {
                    x: mousePos.x - state.latestMouseDownPoint.x + node.latestMouseDownXY.x,
                    y: mousePos.y - state.latestMouseDownPoint.y + node.latestMouseDownXY.y
                  });
                }
                return _extends({}, node);
              }),
              bonds: fragment.bonds.map(function (bond) {
                return _extends({}, bond);
              })
            });

            positionFloatingHydrogens(newFragment);

            var newCs = create2MissingCanvases(newFragment.nodes, newCanvasCache);
            newCanvasCache = [].concat(newCanvasCache, newCs);

            // DETERMINING BOND POLYGONS IN THE new fragment !!!!! new !!!!!
            newFragment.nodes.forEach(function (node, nodeIndex) {
              var c = getCanvas(node, newCanvasCache);
              if (!c) throw 'hmmmm';
              if (node.label) {
                cutBondsWithLabel(nodeIndex, c.filledPixelsData, newFragment.nodes, newFragment.bonds);
              } else {
                cutBondsWithoutLabel(nodeIndex, newFragment.nodes, newFragment.bonds);
              }
            });
            // DETERMINING BOND POLYGONS IN THE new fragment !!!!! new !!!!!

            return newFragment;
          })
        });

        // Because we have moved nodes, we need to determine hovered
        // objects using newObjects
        return _extends({}, state, {
          objects: newObjects,
          pathOfTheObjectBeingHoveredOver: getPathOfTheObjectBeingPointedAt(mousePos, newObjects),
          canvasCache: newCanvasCache
        });
      }
    }

    return _extends({}, state, {
      pathOfTheObjectBeingHoveredOver: getPathOfTheObjectBeingPointedAt(mousePos, state.objects)
    });
  });
  //
  // const mousePos = getMousePos(this.refs.svg, event);
  // console.log(mousePos);
  // console.log(this.state);
  //
  // this.setState(state => {
  //   return {
  //     ...state,
  //     // cursor: (Math.random() < 0.5 ? 'auto' : 'pointer'),
  //     objects: {
  //       ...state.objects,
  //       fragments: state.objects.fragments.map(fragment => ({
  //         ...fragment,
  //         nodes: fragment.nodes.map(node => ({
  //           ...node,
  //           hover: (node.x <= mousePos.x && mousePos.x < node.x + 6 && node.y <= mousePos.y && mousePos.y < node.y + 6)
  //         }))
  //       }))
  //     }
  //   };
  // })
}

var Canvas$1 = function (_Component) {
  inherits(Canvas, _Component);

  function Canvas(props) {
    classCallCheck(this, Canvas);

    var _this = possibleConstructorReturn(this, _Component.call(this, props));

    _this.state = {
      mode: 'VIEWING', // VIEWING, EDITING_LABEL, DRAGGING(?)
      modeData: {},
      // For VIEWING we have an empty modeData object.
      // For EDITING_LABEL we have pathOfTheObjectBeingEdited.
      // For DRAGGING we have an empty modeData object.

      // We render selection & hover halos underneath/below everything else
      pathsOfSelectedObjects: [
        // ['fragments', 1, 'bonds', 4]
      ],
      pathOfTheObjectBeingHoveredOver: undefined,

      latestMouseDownPoint: { // start drag point
        x: undefined,
        y: undefined
      },

      mouseIsDown: false,
      // cursor: 'auto',
      // toolSelected: 'SELECT',
      objects: {
        fragments: [{
          nodes: [],
          bonds: []
        }]
      },
      canvasCache: [
        // {
        //   label: 'C',
        //   focusCharacterRange: [0, atomLine.substring(31,34).trim().length], // one range per line
        //   focusLine: 0, // assuming one line
        //   filledPixelsData: Float32Array(flatten([[0, 0], [0, 1], ...])),
        //   labelOriginXRelativeToFocus: -5.345,
        //   labelOriginYRelativeToFocus: 1.435
        // }
      ],
      minX: 99999, // We will apply Math.min to it.
      minY: 99999, // We will apply Math.min to it.
      maxX: 0, // We will apply Math.max to it.
      maxY: 0 // We will apply Math.max to it.
    };

    _this.pointerUpHandler = _pointerUpHandler.bind(_this);
    _this.pointerDownHandler = _pointerDownHandler.bind(_this);
    _this.pointerMoveHandler = _pointerMoveHandler.bind(_this);
    return _this;
  }

  Canvas.prototype.componentWillUnmount = function componentWillUnmount() {
    this.refs.svg.removeEventListener('touchmove', this.pointerMoveHandler, { passive: false }, false);
  };

  Canvas.prototype.componentDidMount = function componentDidMount() {
    this.refs.svg.addEventListener('touchmove', this.pointerMoveHandler, { passive: false }, false);

    // this.placeMol(
    //   `\n\n\n 25 26  0  0001  0  0  0  0  0999 V2000\n    9.4638   -0.9375    0.0000 S   0  0  0  0  0  0  0  0  0  0  0\n   12.6690   -4.3708    0.0000 Na  0  3  0  0  0  0  0  0  0  0  0\n    5.7264   -3.3446    0.0000 O   0  0  0  0  0  0  0  0  0  0  0\n   11.2628   -4.3455    0.0000 O   0  5  0  0  0  0  0  0  0  0  0\n    9.1344   -5.4097    0.0000 O   0  0  0  0  0  0  0  0  0  0  0\n    4.5989   -2.6732    0.0000 O   0  0  0  0  0  0  0  0  0  0  0\n    8.1969   -2.6732    0.0000 N   0  0  0  0  0  0  0  0  0  0  0\n    5.7264   -0.6715    0.0000 N   0  0  0  0  0  0  0  0  0  0  0\n    8.1969   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    6.8666   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    9.4638   -3.0786    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    6.8666   -2.6732    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n   10.2619   -2.0017    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    9.9325   -4.3455    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n   11.1994   -2.9392    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n   11.1994   -1.0769    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    4.5989   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    3.4333   -0.6715    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    2.3058   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    1.1529   -0.6715    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    2.3058   -2.6732    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    0.0000   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    1.1529   -3.3320    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    0.0000   -2.6732    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    8.1969    0.0000    0.0000 H   0  0  0  0  0  0  0  0  0  0  0\n  1  9  1  0\n  1 13  1  0\n  3 12  2  0\n  4 14  1  0\n  5 14  2  0\n  6 17  2  0\n  7  9  1  0\n  7 11  1  0\n  7 12  1  0\n 10  8  1  1\n  8 17  1  0\n  9 10  1  0\n  9 25  1  6\n 10 12  1  0\n 11 13  1  0\n 11 14  1  6\n 13 15  1  0\n 13 16  1  0\n 17 18  1  0\n 18 19  1  0\n 19 20  2  0\n 19 21  1  0\n 20 22  1  0\n 21 23  2  0\n 22 24  2  0\n 23 24  1  0\nM  CHG  2   2   1   4  -1\nM  END\n`,
    //   this.state
    // );
    this.placeCdxml('<?xml version="1.0" encoding="UTF-8" ?>\n<!DOCTYPE CDXML SYSTEM "http://www.cambridgesoft.com/xml/cdxml.dtd" >\n<CDXML\n CreationProgram="ChemDraw 14.0.0.117"\n Name="Untitled ACS Document 1996-4"\n BoundingBox="118 78.45 239.82 187.74"\n WindowPosition="0 0"\n WindowSize="0 -2147483648"\n WindowIsZoomed="yes"\n FractionalWidths="yes"\n InterpretChemically="yes"\n ShowAtomQuery="yes"\n ShowAtomStereo="no"\n ShowAtomEnhancedStereo="yes"\n ShowAtomNumber="no"\n ShowResidueID="no"\n ShowBondQuery="yes"\n ShowBondRxn="yes"\n ShowBondStereo="no"\n ShowTerminalCarbonLabels="no"\n ShowNonTerminalCarbonLabels="no"\n HideImplicitHydrogens="no"\n LabelFont="90"\n LabelSize="10"\n LabelFace="96"\n CaptionFont="4"\n CaptionSize="12"\n HashSpacing="2.70"\n MarginWidth="2"\n LineWidth="1"\n BoldWidth="4"\n BondLength="28.35"\n BondSpacing="12"\n ChainAngle="120"\n LabelJustification="Auto"\n CaptionJustification="Left"\n AminoAcidTermini="HOH"\n ShowSequenceTermini="yes"\n ShowSequenceBonds="yes"\n ResidueWrapCount="40"\n ResidueBlockCount="10"\n ResidueZigZag="yes"\n NumberResidueBlocks="no"\n PrintMargins="36 36 36 36"\n MacPrintInfo="0003000001200120000000000B6608A0FF84FF880BE309180367052703FC0002000001200120000000000B6608A0000100640064000000010001010100000001270F000100010000000000000000000000000002001901900000000000600000000000000000000100000000000000000000000000000000"\n ChemPropName=""\n ChemPropFormula="Chemical Formula: "\n ChemPropExactMass="Exact Mass: "\n ChemPropMolWt="Molecular Weight: "\n ChemPropMOverZ="m/z: "\n ChemPropAnalysis="Elemental Analysis: "\n ChemPropBoilingPt="Boiling Point: "\n ChemPropMeltingPt="Melting Point: "\n ChemPropCritTemp="Critical Temp: "\n ChemPropCritPres="Critical Pres: "\n ChemPropCritVol="Critical Vol: "\n ChemPropGibbs="Gibbs Energy: "\n ChemPropLogP="Log P: "\n ChemPropMR="MR: "\n ChemPropHenry="Henry&apos;s Law: "\n ChemPropEForm="Heat of Form: "\n ChemProptPSA="tPSA: "\n ChemPropCLogP="CLogP: "\n ChemPropCMR="CMR: "\n ChemPropLogS="LogS: "\n ChemPropPKa="pKa: "\n color="0"\n bgcolor="1"\n><colortable>\n<color r="1" g="1" b="1"/>\n<color r="0" g="0" b="0"/>\n<color r="1" g="0" b="0"/>\n<color r="1" g="1" b="0"/>\n<color r="0" g="1" b="0"/>\n<color r="0" g="1" b="1"/>\n<color r="0" g="0" b="1"/>\n<color r="1" g="0" b="1"/>\n</colortable><fonttable>\n<font id="3" charset="iso-8859-1" name="Arial"/>\n<font id="4" charset="iso-8859-1" name="Times New Roman"/>\n<font id="90" charset="iso-8859-1" name="Helvetica"/>\n</fonttable><page\n id="1472"\n BoundingBox="0 0 540 719.75"\n HeaderPosition="36"\n FooterPosition="36"\n PrintTrimMarks="yes"\n HeightPages="1"\n WidthPages="1"\n><fragment\n id="1435"\n BoundingBox="118 78.45 239.82 187.74"\n Z="26"\n><n\n id="1432"\n p="170.05 132.03"\n Z="27"\n Element="7"\n NumHydrogens="1"\n NeedsClean="yes"\n AS="N"\n><t\n p="166.65 135.71"\n BoundingBox="166.65 127.83 180.25 135.71"\n LabelJustification="Left"\n LabelAlignment="Left"\n><s font="3" size="9.4" face="96">NH</s></t></n><n\n id="1434"\n p="170.05 103.68"\n Z="28"\n Element="7"\n NumHydrogens="1"\n NeedsClean="yes"\n AS="N"\n><t\n p="166.65 107.36"\n BoundingBox="166.65 99.48 180.25 107.36"\n LabelJustification="Left"\n LabelAlignment="Left"\n><s font="3" size="9.4" face="96">NH</s></t></n><n\n id="1436"\n p="141.70 103.68"\n Z="29"\n/><n\n id="1438"\n p="141.70 132.03"\n Z="30"\n AS="N"\n/><n\n id="1444"\n p="194.60 117.86"\n Z="34"\n Element="7"\n NumHydrogens="1"\n NeedsClean="yes"\n AS="N"\n><t\n p="198 121.53"\n BoundingBox="184.40 113.66 198 121.53"\n LabelJustification="Right"\n Justification="Right"\n LabelAlignment="Right"\n><s font="3" size="9.4" face="96">NH</s></t></n><n\n id="1446"\n p="208.78 93.31"\n Z="35"\n/><n\n id="1448"\n p="184.22 79.13"\n Z="36"\n AS="N"\n/><n\n id="1454"\n p="184.22 156.59"\n Z="40"\n/><n\n id="1456"\n p="208.78 142.41"\n Z="41"\n AS="N"\n/><n\n id="1462"\n p="121.65 83.64"\n Z="45"\n Element="8"\n NumHydrogens="0"\n NeedsClean="yes"\n AS="N"\n><t\n p="118 87.31"\n BoundingBox="118 79.32 125.31 87.41"\n LabelJustification="Left"\n><s font="3" size="9.4" face="96">O</s></t></n><n\n id="1465"\n p="176.89 183.97"\n Z="47"\n Element="8"\n NumHydrogens="0"\n NeedsClean="yes"\n AS="N"\n><t\n p="173.24 187.64"\n BoundingBox="173.24 179.65 180.55 187.74"\n LabelJustification="Left"\n><s font="3" size="9.4" face="96">O</s></t></n><n\n id="1468"\n p="236.16 85.97"\n Z="49"\n Element="8"\n NumHydrogens="0"\n NeedsClean="yes"\n AS="N"\n><t\n p="232.51 89.64"\n BoundingBox="232.51 81.65 239.82 89.74"\n LabelJustification="Left"\n><s font="3" size="9.4" face="96">O</s></t></n><b\n id="1441"\n Z="31"\n B="1434"\n E="1436"\n BS="N"\n/><b\n id="1442"\n Z="32"\n B="1436"\n E="1438"\n BS="N"\n/><b\n id="1443"\n Z="33"\n B="1438"\n E="1432"\n BS="N"\n/><b\n id="1451"\n Z="37"\n B="1444"\n E="1446"\n BS="N"\n/><b\n id="1452"\n Z="38"\n B="1446"\n E="1448"\n BS="N"\n/><b\n id="1453"\n Z="39"\n B="1448"\n E="1434"\n BS="N"\n/><b\n id="1459"\n Z="42"\n B="1432"\n E="1454"\n BS="N"\n/><b\n id="1460"\n Z="43"\n B="1454"\n E="1456"\n BS="N"\n/><b\n id="1461"\n Z="44"\n B="1456"\n E="1444"\n BS="N"\n/><b\n id="1464"\n Z="46"\n B="1436"\n E="1462"\n Order="2"\n/><b\n id="1467"\n Z="48"\n B="1454"\n E="1465"\n Order="2"\n/><b\n id="1470"\n Z="50"\n B="1446"\n E="1468"\n Order="2"\n/></fragment></page></CDXML>', this.state);
    // this.setState(state => {
    //   return {
    //     ...state,
    //     pathsOfSelectedObjects: [
    //       ['fragments', 1, 'bonds', 4],
    //       ['fragments', 1, 'bonds', 3]
    //     ]
    //   };
    // });
  };

  Canvas.prototype.placeMol = function placeMol(mol, state) {

    var fragment = readMol(mol);

    var newCs = create2MissingCanvases(fragment.nodes, state.canvasCache);

    // DETERMINING BOND POLYGONS IN THE fragment
    fragment.nodes.forEach(function (node, nodeIndex) {
      var c = getCanvas(node, [].concat(state.canvasCache, newCs));
      if (!c) throw 'hmmmm';
      if (node.label) {
        cutBondsWithLabel(nodeIndex, c.filledPixelsData, fragment.nodes, fragment.bonds);
      } else {
        cutBondsWithoutLabel(nodeIndex, fragment.nodes, fragment.bonds);
      }
    });
    // DETERMINING BOND POLYGONS IN THE fragment


    var newMinX = state.minX;
    var newMinY = state.minY;
    var newMaxX = state.maxX;
    var newMaxY = state.maxY;
    fragment.nodes.forEach(function (node) {
      var c = getCanvas(node, [].concat(state.canvasCache, newCs));
      if (!c) throw 'hmmmm';
      newMinX = Math.min(newMinX, node.x + c.minX - CONSTANTS.SVG_PADDING);
      newMinY = Math.min(newMinY, node.y + c.minY - CONSTANTS.SVG_PADDING);
      newMaxX = Math.max(newMaxX, node.x + c.maxX + CONSTANTS.SVG_PADDING);
      newMaxY = Math.max(newMaxY, node.y + c.maxY + CONSTANTS.SVG_PADDING);
    });

    this.setState(function (state) {
      return _extends({}, state, {
        objects: _extends({}, state.objects, {
          fragments: [].concat(state.objects.fragments, [fragment])
        }),
        canvasCache: [].concat(state.canvasCache, newCs),
        minX: newMinX,
        minY: newMinY,
        maxX: newMaxX,
        maxY: newMaxY
      });
    });
  };

  Canvas.prototype.placeCdxml = function placeCdxml(cdxml, state) {

    var fragment = readCdxml(cdxml);

    var newCs = create2MissingCanvases(fragment.nodes, state.canvasCache);

    // DETERMINING BOND POLYGONS IN THE fragment
    fragment.nodes.forEach(function (node, nodeIndex) {
      var c = getCanvas(node, [].concat(state.canvasCache, newCs));
      if (!c) throw 'hmmmm';
      if (node.label) {
        cutBondsWithLabel(nodeIndex, c.filledPixelsData, fragment.nodes, fragment.bonds);
      } else {
        cutBondsWithoutLabel(nodeIndex, fragment.nodes, fragment.bonds);
      }
    });
    // DETERMINING BOND POLYGONS IN THE fragment


    var newMinX = state.minX;
    var newMinY = state.minY;
    var newMaxX = state.maxX;
    var newMaxY = state.maxY;
    fragment.nodes.forEach(function (node) {
      var c = getCanvas(node, [].concat(state.canvasCache, newCs));
      if (!c) throw 'hmmmm';
      newMinX = Math.min(newMinX, node.x + c.minX - CONSTANTS.SVG_PADDING);
      newMinY = Math.min(newMinY, node.y + c.minY - CONSTANTS.SVG_PADDING);
      newMaxX = Math.max(newMaxX, node.x + c.maxX + CONSTANTS.SVG_PADDING);
      newMaxY = Math.max(newMaxY, node.y + c.maxY + CONSTANTS.SVG_PADDING);
    });

    this.setState(function (state) {
      return _extends({}, state, {
        objects: _extends({}, state.objects, {
          fragments: [].concat(state.objects.fragments, [fragment])
        }),
        canvasCache: [].concat(state.canvasCache, newCs),
        minX: newMinX,
        minY: newMinY,
        maxX: newMaxX,
        maxY: newMaxY
      });
    });
  };

  Canvas.prototype.setLabel = function setLabel(fragmentIndex, nodeIndex, newLabel, state) {

    var oldFragment = state.objects.fragments[fragmentIndex];
    var newNodes = [].concat(oldFragment.nodes);
    newNodes[nodeIndex] = _extends({}, newNodes[nodeIndex], {
      label: newLabel
    });
    var newFragment = _extends({}, oldFragment, {
      nodes: newNodes
    });

    var newCs = create2MissingCanvases(newFragment.nodes, state.canvasCache);

    // DETERMINING BOND POLYGONS IN THE newFragment !!!! new !!!!!
    newFragment.nodes.forEach(function (node, nodeIndex) {
      var c = getCanvas(node, [].concat(state.canvasCache, newCs));
      if (!c) throw 'hmmmm';
      if (node.label) {
        cutBondsWithLabel(nodeIndex, c.filledPixelsData, newFragment.nodes, newFragment.bonds);
      } else {
        cutBondsWithoutLabel(nodeIndex, newFragment.nodes, newFragment.bonds);
      }
    });
    // DETERMINING BOND POLYGONS IN THE newFragment !!!! new !!!!!


    var newFragments = state.objects.fragments;
    newFragments[fragmentIndex] = newFragment;
    this.setState(function (state) {
      return _extends({}, state, {
        objects: _extends({}, state.objects, {
          fragments: newFragments
        }),
        canvasCache: [].concat(state.canvasCache, newCs)
      });
    });
  };

  Canvas.prototype.renderSelectedObjectHalos = function renderSelectedObjectHalos(state) {

    // Returning an array of two elements: selection halos and hover halo
    return [state.pathsOfSelectedObjects.map(function (objectPath, objectPathIndex) {
      return getHaloFromObjectPath(objectPath, objectPathIndex, false, state);
    }), state.pathOfTheObjectBeingHoveredOver && getHaloFromObjectPath(state.pathOfTheObjectBeingHoveredOver, state.pathsOfSelectedObjects.length, true, state)];
  };

  Canvas.prototype.render = function render() {
    var _this2 = this;

    var state = this.state;

    return React__default.createElement(
      'div',
      { style: {
          position: 'relative',
          width: state.maxX - state.minX + 'px',
          height: state.maxY - state.minY + 'px',
          display: 'inline-block'
          //opacity: 0.05
        } },
      React__default.createElement(
        'svg',
        {
          viewBox: state.minX + ' ' + state.minY + ' ' + (state.maxX - state.minX) + ' ' + (state.maxY - state.minY),
          style: {
            position: 'absolute',
            top: '0',
            left: '0',
            cursor: state.pathOfTheObjectBeingHoveredOver ? 'pointer' : 'auto',
            width: state.maxX - state.minX + 'px',
            height: state.maxY - state.minY + 'px'
          },
          ref: 'svg',
          width: state.maxX - state.minX,
          height: state.maxY - state.minY,
          onMouseDown: this.pointerDownHandler,
          onMouseUp: this.pointerUpHandler,
          onMouseMove: this.pointerMoveHandler,
          onTouchStart: this.pointerDownHandler,
          onTouchEnd: this.pointerUpHandler
        },
        this.renderSelectedObjectHalos(state),
        state.objects.fragments.map(function (fragment, fragmentIndex) {
          return React__default.createElement(Fragment, {
            key: fragmentIndex,
            canvasCache: state.canvasCache,
            fragment: fragment
          });
        })
      ),
      state.objects.fragments.map(function (fragment, fragmentIndex) {
        return React__default.createElement(
          'div',
          { key: fragmentIndex, style: { position: 'absolute', top: '0', left: '0' } },
          fragment.nodes.map(function (node, nodeIndex) {
            if (!node.beingEdited) return;
            var c = getCanvas(node, state.canvasCache);
            if (!c) return;
            return React__default.createElement(Label, {
              key: nodeIndex,
              origin: {
                x: node.x + c.labelOriginXRelativeToFocus,
                y: node.y + c.labelOriginYRelativeToFocus
              },
              onChange: function onChange(newLabel) {
                _this2.setLabel(fragmentIndex, nodeIndex, newLabel, state);
              },
              value: node.label
            });
          })
        );
      })
    );
  };

  return Canvas;
}(React.Component);

return Canvas$1;

})));
