'use strict';

exports.__esModule = true;

exports.default = function (nodes, canvasCache) {
  var newCs = [];

  nodes.forEach(function (node) {

    var existingC = (0, _helpers.getCanvas)(node, [].concat(canvasCache, newCs));
    if (existingC) return;

    var canvasDomElement = document.createElement('CANVAS');
    var ctx = canvasDomElement.getContext('2d');
    ctx.font = _constants2.default.FONT_SIZE + 'px sans-serif';
    ctx.strokeStyle = 'red';
    ctx.lineWidth = _constants2.default.LABEL_BOND_SPACING;
    ctx.lineJoin = 'round';
    // const nodeLabelLines = node.label.split('\n');

    // Assuming one line. (We call it label for now.)
    // Assuming no kerning or ligatures.
    // Y is directed downwards.
    // We will have one focus character range per line; center of the range is the center of the line.
    // Origin is the point we specify when using ctx.fillText/strokeText (left bottom point of text).
    // Bottom of the label is the baseline. Height is font size.
    var labelOriginXRelativeToFocus = -ctx.measureText(node.label.substring(0, node.focusCharacterRange[0])).width - ctx.measureText(node.label.substring(node.focusCharacterRange[0], node.focusCharacterRange[1])).width / 2;
    var labelOriginYRelativeToFocus = _constants2.default.FONT_SIZE / 3;
    var labelWidth = ctx.measureText(node.label).width;
    var labelHeight = _constants2.default.FONT_SIZE;

    // Adding padding to our canvases for "ygp..." protrusions etc.
    var leftCanvasBorderXRelativeToFocus = labelOriginXRelativeToFocus - _constants2.default.ADDED_MARGIN_FOR_CANVASES;
    var rightCanvasBorderXRelativeToFocus = labelOriginXRelativeToFocus + labelWidth + _constants2.default.ADDED_MARGIN_FOR_CANVASES;
    var topCanvasBorderYRelativeToFocus = labelOriginYRelativeToFocus - labelHeight - _constants2.default.ADDED_MARGIN_FOR_CANVASES;
    var bottomCanvasBorderYRelativeToFocus = labelOriginYRelativeToFocus + _constants2.default.ADDED_MARGIN_FOR_CANVASES;

    // const textWidth = Math.max(...nodeLabelLines.map(labelLine => ctx.measureText(labelLine).width));
    // console.log(textWidth);
    // const textHeight = nodeLabelLines.length * CONSTANTS.FONT_SIZE;
    canvasDomElement.width = rightCanvasBorderXRelativeToFocus - leftCanvasBorderXRelativeToFocus;
    canvasDomElement.height = bottomCanvasBorderYRelativeToFocus - topCanvasBorderYRelativeToFocus;

    // Our ctx got reset after we resized the canvas.
    ctx.font = _constants2.default.FONT_SIZE + 'px sans-serif';
    ctx.strokeStyle = 'red';
    ctx.lineWidth = _constants2.default.LABEL_BOND_SPACING;
    ctx.lineJoin = 'round';

    // nodeLabelLines.forEach((labelLine, labelLineIndex) => {
    //   ctx.strokeText(labelLine, ADDED_MARGIN_FOR_CANVASES, ADDED_MARGIN_FOR_CANVASES + (labelLineIndex + 1) * CONSTANTS.FONT_SIZE);
    // });
    ctx.strokeText(node.label, labelOriginXRelativeToFocus - leftCanvasBorderXRelativeToFocus, labelOriginYRelativeToFocus - topCanvasBorderYRelativeToFocus);

    var pixelData = ctx.getImageData(0, 0, canvasDomElement.width, canvasDomElement.height).data;

    var filledPixels = [];
    for (var x = 0; x < canvasDomElement.width; ++x) {
      for (var y = 0; y < canvasDomElement.height; ++y) {
        if (pixelData[y * canvasDomElement.width * 4 + x * 4 + 3] > 0) {
          // max for alpha is 255?
          filledPixels.push([x + leftCanvasBorderXRelativeToFocus + 0.5, y + topCanvasBorderYRelativeToFocus + 0.5]);
        }
      }
    }

    // The Y coordinate line is directed downwards.
    var minX = 99999; // We will apply Math.min to it.
    var minY = 99999; // We will apply Math.min to it.
    var maxX = 0; // We will apply Math.max to it.
    var maxY = 0; // // We will apply Math.max to it.

    filledPixels.forEach(function (filledPixel) {
      minX = Math.min(minX, filledPixel[0]);
      minY = Math.min(minY, filledPixel[1]);
      maxX = Math.max(maxX, filledPixel[0]);
      maxY = Math.max(maxY, filledPixel[1]);
    });

    var filledPixelsFlat = filledPixels.reduce(function (acc, filledPixel) {
      acc.push(filledPixel[0]);
      acc.push(filledPixel[1]);
      return acc;
    }, []);
    var filledPixelsData = new Float32Array(filledPixelsFlat);

    var c = {
      label: node.label,
      focusCharacterRange: node.focusCharacterRange, // one range per line
      focusLine: node.focusLine,
      filledPixels: filledPixels,
      filledPixelsData: filledPixelsData, // origin here is the focus point which is in the middle of focus line (both in X and in Y coordinates)
      labelOriginXRelativeToFocus: labelOriginXRelativeToFocus,
      labelOriginYRelativeToFocus: labelOriginYRelativeToFocus,
      minX: minX,
      minY: minY,
      maxX: maxX,
      maxY: maxY
    };

    // this.setState(state => {
    //   // if (state.canvasCache.some(c => {
    //   //   return c.label === node.label;
    //   // })) {
    //   //   return state;
    //   // }
    //   return {
    //     ...state,
    //     canvasCache: [...state.canvasCache, c]
    //   };
    // });


    newCs.push(c);
    //return c;
  });
  return newCs;
};

var _helpers = require('./helpers');

var _constants = require('./constants');

var _constants2 = _interopRequireDefault(_constants);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }