'use strict';

exports.__esModule = true;
exports.pointIsToTheLeft = pointIsToTheLeft;
exports.pointIsInside = pointIsInside;
exports.polygonOverlapsImg = polygonOverlapsImg;
exports.createPoint = createPoint;
exports.addPoints = addPoints;
exports.subtractPoints = subtractPoints;
exports.scalePoint = scalePoint;
exports.slidePoint = slidePoint;
exports.pointToPathString = pointToPathString;
exports.projectAPointOntoALine = projectAPointOntoALine;
exports.distanceBetweenTwoPoints = distanceBetweenTwoPoints;
exports.getBondsInvolved = getBondsInvolved;
exports.getBondD = getBondD;
exports.getVectorLength = getVectorLength;
exports.getVectorAngle = getVectorAngle;
exports.reversePoint = reversePoint;
exports.intersection = intersection;
exports.cutBondsWithLabel = cutBondsWithLabel;
exports.cutBondsWithoutLabel = cutBondsWithoutLabel;
exports.getCanvas = getCanvas;
exports.getPathOfTheObjectBeingPointedAt = getPathOfTheObjectBeingPointedAt;
exports.arraysAreEqual = arraysAreEqual;
exports.arrayOfArraysIncludesArray = arrayOfArraysIncludesArray;
exports.getHaloFromObjectPath = getHaloFromObjectPath;

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _constants = require('./constants');

var _constants2 = _interopRequireDefault(_constants);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function pointIsToTheLeft(c, //the point
a, //line point from
b //line point to
) {
  return (b.x - a.x) * (c.y - a.y) - (b.y - a.y) * (c.x - a.x) > 0;
}

function pointIsInside(c, //the point
polygonPoints //array of polygon points
) {
  for (var i = 0; i < polygonPoints.length - 1; ++i) {
    var polygonPoint1 = polygonPoints[i];
    var polygonPoint2 = polygonPoints[i + 1];
    if (!pointIsToTheLeft(c, polygonPoint1, polygonPoint2)) return false;
  }

  return true;
}

function polygonOverlapsImg(polygonPoints, filledPixelsData) {

  // const data1 = new Float32Array([11, 22.5, 33]);
  // const mem1 = window._malloc(data1 * Float32Array.BYTES_PER_ELEMENT);
  // window.HEAPF32.set(data1, mem1 / Float32Array.BYTES_PER_ELEMENT);
  // const data2 = new Float32Array([44, -55.75, 66]);
  // const mem2 = window._malloc(data2 * Float32Array.BYTES_PER_ELEMENT);
  // window.HEAPF32.set(data2, mem2 / Float32Array.BYTES_PER_ELEMENT);
  // console.log(window._testMyFloats(mem1, mem2));
  // window._free(mem1);
  // window._free(mem2);

  var polygonPointsFlat = polygonPoints.reduce(function (acc, point) {
    // console.log(point);
    acc.push(point.x);
    acc.push(point.y);
    return acc;
  }, []);
  var polygonPointsData = new Float32Array(polygonPointsFlat);
  // console.log(polygonPointsData);
  //console.log(polygonPointsData[1]);
  var polygonPointsMem = window._malloc(polygonPoints.length * 2 * Float32Array.BYTES_PER_ELEMENT);
  //window.HEAPF32.set(polygonPointsData, polygonPointsMem / Float32Array.BYTES_PER_ELEMENT);

  // const filledPixelsFlat = filledPixels.reduce((acc, filledPixel) => {
  //   acc.push(filledPixel[0]);
  //   acc.push(filledPixel[1]);
  //   return acc;
  // }, []);
  //const filledPixelsData = new Float32Array(filledPixelsFlat);
  var filledPixelsMem = window._malloc(filledPixelsData.length * Float32Array.BYTES_PER_ELEMENT);
  //window.HEAPF32.set(filledPixelsData, filledPixelsMem / Float32Array.BYTES_PER_ELEMENT);

  // console.log(filledPixelsData.length, filledPixels.length);
  // console.log(filledPixelsMem);

  var toReturn = window._polygonOverlapsImg(polygonPointsData, polygonPoints.length, filledPixelsData, filledPixelsData.length / 2);
  window._free(polygonPointsMem);
  window._free(filledPixelsMem);
  // console.log(toReturn);
  // console.log(polygonPointsData[polygonPointsData.length - 1]);
  return toReturn;

  // float *polygonPoints,
  // int polygonPointsLength,
  // float *filledPixels,
  // int filledPixelsLength


  // for (var r = 0; r < filledPixels.length; ++r) {
  // 	var p = filledPixels[r];
  // 	if (pointIsInside(createPoint(p[0] + 0.5, p[1] + 0.5), polygonPoints)) return true;
  // }
  // return false;
}

function createPoint(x, y) {
  return { x: x, y: y };
}

function addPoints(p1, p2) {
  return {
    x: p1.x + p2.x,
    y: p1.y + p2.y
  };
}

function subtractPoints(p1, p2) {
  return {
    x: p1.x - p2.x,
    y: p1.y - p2.y
  };
}

function scalePoint(p, k) {
  return {
    x: k * p.x,
    y: k * p.y
  };
}

function slidePoint(p1, p2, t) {
  return createPoint(p1.x + t * (p2.x - p1.x), p1.y + t * (p2.y - p1.y));
}

function pointToPathString(p) {
  return p.x + ' ' + p.y;
}

function projectAPointOntoALine(A, B, C) {
  // A and B are the point that define the line.
  // C is the point that we a projecting onto the line.
  // AB is the vector from A to B.
  // ABperp is the AB vector rotated by 90 degrees counterclockwise.
  // u is the slider for AB vector if we attach it to A.
  // v is the slider for ABperp vector if we attach it to C.
  // u and v determine the coordinates of the projection point.
  // P is the projection point.

  var AB = {
    x: B.x - A.x,
    y: B.y - A.y
  };
  var ABperp = {
    x: A.y - B.y,
    y: B.x - A.x
  };

  // A + u * AB == C + v * ABperp.

  // A.x + u * AB.x == C.x + v * ABperp.x
  // A.y + u * AB.y == C.y + v * ABperp.y

  // u == (C.x + v * ABperp.x - A.x) / AB.x
  // u == (C.y + v * ABperp.y - A.y) / AB.y

  // (C.x + v * ABperp.x - A.x) / AB.x == (C.y + v * ABperp.y - A.y) / AB.y

  // (C.x + v * ABperp.x - A.x) * AB.y == (C.y + v * ABperp.y - A.y) * AB.x

  // C.x * AB.y + v * ABperp.x * AB.y - A.x * AB.y == C.y * AB.x + v * ABperp.y * AB.x - A.y * AB.x

  // v * ABperp.x * AB.y - v * ABperp.y * AB.x == C.y * AB.x - A.y * AB.x - C.x * AB.y + A.x * AB.y

  // v = (C.y * AB.x - A.y * AB.x - C.x * AB.y + A.x * AB.y) / (ABperp.x * AB.y - ABperp.y * AB.x)

  var v = (C.y * AB.x - A.y * AB.x - C.x * AB.y + A.x * AB.y) / (ABperp.x * AB.y - ABperp.y * AB.x);
  var u = AB.x ? (C.x + v * ABperp.x - A.x) / AB.x : (C.y + v * ABperp.y - A.y) / AB.y;
  var P = {
    x: A.x + u * AB.x,
    y: A.y + u * AB.y
  };
  return { u: u, v: v, P: P };
}

function distanceBetweenTwoPoints(A, B) {
  var d = subtractPoints(B, A);
  return Math.sqrt(d.x * d.x + d.y * d.y);
}

function getBondsInvolved(nodeIndex, fragmentBonds) {
  return fragmentBonds.filter(function (bond) {
    return bond.firstNodeIndex === nodeIndex || bond.secondNodeIndex === nodeIndex;
  });
}

function getBondD(bond, fragmentNodes) {
  var firstNode = fragmentNodes[bond.firstNodeIndex];
  var secondNode = fragmentNodes[bond.secondNodeIndex];
  return subtractPoints(secondNode, firstNode);
}

function getVectorLength(d) {
  return Math.sqrt(d.x * d.x + d.y * d.y);
}

function getVectorAngle(d) {
  return Math.atan2(d.y, d.x);
}

function reversePoint(p) {
  return {
    x: -p.x,
    y: -p.y
  };
}

// t is the proportion of segment1 to the intersection point
function intersection(segment1FirstPointCoordinates, segment1SecondPointCoordinates, //
segment2FirstPointCoordinates, segment2SecondPointCoordinates //
) {

  var p = segment1FirstPointCoordinates;
  var q = segment2FirstPointCoordinates;

  var r = {
    x: segment1SecondPointCoordinates.x - segment1FirstPointCoordinates.x,
    y: segment1SecondPointCoordinates.y - segment1FirstPointCoordinates.y
  };
  var s = {
    x: segment2SecondPointCoordinates.x - segment2FirstPointCoordinates.x,
    y: segment2SecondPointCoordinates.y - segment2FirstPointCoordinates.y
  };
  //points p and q
  //vectors r and s from those points

  //intersection point is at (p + t*r) or (q + u*s)
  //p.x + t*r.x = q.x + u*s.x
  //p.y + t*r.y = q.y + u*s.y <=> u = (p.y + t*r.y - q.y) / s.y
  //
  // =>
  //
  //p.x + t*r.x = q.x + [(p.y + t*r.y - q.y) / s.y * s.x] = q.x + p.y*s.x/s.y + t*r.y*s.x/s.y - q.y*s.x/s.y
  // <=>
  //t = (q.x + p.y*s.x/s.y - q.y*s.x/s.y - p.x) / (r.x - r.y*s.x/s.y)

  if (Math.abs(s.y) < 0.0001) {
    //segment 2 is horizontal
    return ((segment2SecondPointCoordinates.y + segment2FirstPointCoordinates.y) / 2 - segment1FirstPointCoordinates.y) / r.y;
  } else {
    return (q.x + p.y * s.x / s.y - q.y * s.x / s.y - p.x) / (r.x - r.y * s.x / s.y);
  }
}

function cutBondsWithLabel(nodeIndex, filledPixelsData, fragmentNodes, fragmentBonds) {
  var bondsInvolved = getBondsInvolved(nodeIndex, fragmentBonds);

  bondsInvolved.forEach(function (bond) {

    var firstNode = fragmentNodes[bond.firstNodeIndex];
    var secondNode = fragmentNodes[bond.secondNodeIndex];

    var d = getBondD(bond, fragmentNodes);
    var length = getVectorLength(d);
    // If horizontal unit vector is along X, vertical unit vector is directed upwards.
    var bondVerticalUnitVector = createPoint(0 - d.y / length, d.x / length);
    var tStep = 1 / length / 2;

    if (bond.firstNodeIndex === nodeIndex) {

      switch (bond.type) {
        case 1:
        default:

          bond.firstT = 0;

          for (var filledPixelIndex = 0; filledPixelIndex < filledPixelsData.length; filledPixelIndex += 2) {

            var C = {
              x: filledPixelsData[filledPixelIndex],
              y: filledPixelsData[filledPixelIndex + 1]
            };

            var _projectAPointOntoALi = projectAPointOntoALine({ x: 0, y: 0 }, subtractPoints(secondNode, firstNode), C),
                P = _projectAPointOntoALi.P,
                u = _projectAPointOntoALi.u;

            if (distanceBetweenTwoPoints(P, C) < _constants2.default.BOND_BAR_WIDTH / 2) {
              bond.firstT = Math.max(bond.firstT, u);
            }
          }

          // bond.firstT = 0;
          //
          // var t = 0;
          // for (t = 0; t <= 1; t+=tStep) {
          //
          //
          //
          //
          //   if (!polygonOverlapsImg([
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), firstNode),
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), firstNode),
          //     subtractPoints(
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       firstNode
          //     ),
          //     subtractPoints(
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       firstNode
          //     ),
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), firstNode)
          //   ], filledPixelsData)) break;
          //
          // }
          //
          // bond.firstT = t;

          break;

        // case 2:

        // var t = 0;
        // for (t = 0; t <= 1; t+=0.005) {
        //
        //
        //
        //
        //   if (!polygonOverlapsImg(ctx, [
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //       secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //       t
        //     ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //       secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //       t
        //     ).added(bondHorizontalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //       secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //       t
        //     ).added(bondHorizontalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //       secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //       t
        //     ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
        //   ], canvas, thisAtom)
        //
        //   &&
        //
        //   !polygonOverlapsImg(ctx, [
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //       secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //       t
        //     ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //       secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //       t
        //     ).added(bondHorizontalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //       secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //       t
        //     ).added(bondHorizontalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     slide(
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //       secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //       t
        //     ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //     secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
        //   ], canvas, thisAtom)) break;
        //
        // }
        //
        // bond.firstT = t;
        //
        // break;

      }
    } else {
      //
      //
      //
      switch (bond.type) {
        case 1:
        default:

          bond.secondT = 1;

          for (var _filledPixelIndex = 0; _filledPixelIndex < filledPixelsData.length; _filledPixelIndex += 2) {

            var _C = {
              x: filledPixelsData[_filledPixelIndex],
              y: filledPixelsData[_filledPixelIndex + 1]
            };

            var _projectAPointOntoALi2 = projectAPointOntoALine(subtractPoints(firstNode, secondNode), { x: 0, y: 0 }, _C),
                P = _projectAPointOntoALi2.P,
                u = _projectAPointOntoALi2.u;

            if (distanceBetweenTwoPoints(P, _C) < _constants2.default.BOND_BAR_WIDTH / 2) {
              bond.secondT = Math.min(bond.secondT, u);
            }
          }
          //
          // bond.secondT = 1;
          // var t = 1;
          // for (t = 1; t >= 0; t-=tStep) {
          //
          //
          //   // if (!polygonOverlapsImg(ctx, [
          //   //   firstNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   firstNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   slide(
          //   //     firstNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     secondNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     t
          //   //   ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   slide(
          //   //     firstNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     secondNode.added(bondVerticalUnitVector.scaled(-CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     t
          //   //   ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   slide(
          //   //     firstNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     secondNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     t
          //   //   ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
          //   //   slide(
          //   //     firstNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     secondNode.added(bondVerticalUnitVector.scaled(CONSTANTS.BOND_BAR_WIDTH/2)),
          //   //     t
          //   //   ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
          //   // ], canvas, thisAtom)) break;
          //
          //   if (!polygonOverlapsImg([
          //     subtractPoints(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       secondNode
          //     ),
          //     subtractPoints(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       secondNode
          //     ),
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), secondNode),
          //     subtractPoints(slidePoint(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       t
          //     ), secondNode),
          //     subtractPoints(
          //       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
          //       secondNode
          //     )
          //   ], filledPixelsData)) break;
          //
          // }
          //
          // bond.secondT = t;

          break;
        //   case 2:
        //   var t = 1;
        //   for (t = 1; t >= 0; t-=0.005) {
        //
        //
        //     if (!polygonOverlapsImg(ctx, [
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //         secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //         t
        //       ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //         secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2)),
        //         t
        //       ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //         secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //         t
        //       ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //         secondNode.added(bondVerticalUnitVector.scaled(interBondBarGapWidth/2+CONSTANTS.BOND_BAR_WIDTH)),
        //         t
        //       ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
        //     ], canvas, thisAtom)
        //
        //     &&
        //
        //     !polygonOverlapsImg(ctx, [
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //         secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //         t
        //       ).added(bondVerticalUnitVector.scaled(-bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //         secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2-CONSTANTS.BOND_BAR_WIDTH)),
        //         t
        //       ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //         secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //         t
        //       ).added(bondHorizontalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates),
        //       slide(
        //         firstNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //         secondNode.added(bondVerticalUnitVector.scaled(-interBondBarGapWidth/2)),
        //         t
        //       ).added(bondVerticalUnitVector.scaled(bondBarTransparentBorderWidth)).subtracted(atomCoordinates)
        //     ], canvas, thisAtom)) break;
        //
        //   }
        //
        //   bond.secondT = t;
        //
        //   break;
      }
      //
      //
      //
      // //barCoordinatesArray.secondUpT = t;
      // //barCoordinatesArray.secondDownT = t;
    }
  });
}

function cutBondsWithoutLabel(nodeIndex, fragmentNodes, fragmentBonds) {

  var bondsInvolved = getBondsInvolved(nodeIndex, fragmentBonds);

  bondsInvolved.sort(function (bond1, bond2) {

    var d1 = getBondD(bond1, fragmentNodes);
    var d2 = getBondD(bond2, fragmentNodes);

    var bond1IsDirectedOutwards = bond1.firstNodeIndex === nodeIndex;
    var bond2IsDirectedOutwards = bond2.firstNodeIndex === nodeIndex;

    var angle1 = void 0,
        angle2 = void 0;
    if (bond1IsDirectedOutwards) {
      if (bond2IsDirectedOutwards) {
        //bond 1 and bond 2 are directed outwards
        angle1 = getVectorAngle(d1);
        angle2 = getVectorAngle(d2);
      } else {
        //bond 1 is directed outwards and bond 2 is directed inwards
        angle1 = getVectorAngle(d1);
        angle2 = getVectorAngle(reversePoint(d2));
      }
    } else {
      if (bond2IsDirectedOutwards) {
        //bond 1 is directed inwards and bond 2 is directed outwards
        angle1 = getVectorAngle(reversePoint(d1));
        angle2 = getVectorAngle(d2);
      } else {
        //bond 1 and bond 2 are directed inwards
        angle1 = getVectorAngle(reversePoint(d1));
        angle2 = getVectorAngle(reversePoint(d2));
      }
    }

    // bond1 should come first if angle1 < angle2

    return angle1 - angle2;
  });

  if (bondsInvolved.length === 1) return;

  for (var s = 0; s < bondsInvolved.length; s++) {

    var bond1 = bondsInvolved[s];
    var bond2 = bondsInvolved[(s + 1) % bondsInvolved.length];

    var bond1IsDirectedOutwards = bond1.firstNodeIndex === nodeIndex;
    var bond2IsDirectedOutwards = bond2.firstNodeIndex === nodeIndex;

    var d1 = getBondD(bond1, fragmentNodes);
    var length1 = getVectorLength(d1);
    var bond1VerticalUnitVector = createPoint(-d1.y / length1, d1.x / length1);

    var d2 = getBondD(bond2, fragmentNodes);
    var length2 = getVectorLength(d2);
    var bond2VerticalUnitVector = createPoint(-d2.y / length2, d2.x / length2);

    var bond1FirstNode = fragmentNodes[bond1.firstNodeIndex];
    var bond1SecondNode = fragmentNodes[bond1.secondNodeIndex];

    var bond2FirstNode = fragmentNodes[bond2.firstNodeIndex];
    var bond2SecondNode = fragmentNodes[bond2.secondNodeIndex];

    var bond1FirstUp = addPoints(bond1FirstNode, scalePoint(bond1VerticalUnitVector, _constants2.default.BOND_BAR_WIDTH / 2));
    var bond1FirstDown = addPoints(bond1FirstNode, scalePoint(bond1VerticalUnitVector, -_constants2.default.BOND_BAR_WIDTH / 2));

    var bond1SecondUp = addPoints(bond1SecondNode, scalePoint(bond1VerticalUnitVector, _constants2.default.BOND_BAR_WIDTH / 2));
    var bond1SecondDown = addPoints(bond1SecondNode, scalePoint(bond1VerticalUnitVector, -_constants2.default.BOND_BAR_WIDTH / 2));

    var bond2FirstUp = addPoints(bond2FirstNode, scalePoint(bond2VerticalUnitVector, _constants2.default.BOND_BAR_WIDTH / 2));
    var bond2FirstDown = addPoints(bond2FirstNode, scalePoint(bond2VerticalUnitVector, -_constants2.default.BOND_BAR_WIDTH / 2));

    var bond2SecondUp = addPoints(bond2SecondNode, scalePoint(bond2VerticalUnitVector, _constants2.default.BOND_BAR_WIDTH / 2));
    var bond2SecondDown = addPoints(bond2SecondNode, scalePoint(bond2VerticalUnitVector, -_constants2.default.BOND_BAR_WIDTH / 2));

    if (bond1IsDirectedOutwards) {
      if (bond2IsDirectedOutwards) {
        //bond 1 and bond 2 are directed outwards
        bond1.firstUpT = intersection(bond1FirstUp, bond1SecondUp, bond2FirstDown, bond2SecondDown);
        bond2.firstDownT = intersection(bond2FirstDown, bond2SecondDown, bond1FirstUp, bond1SecondUp);
      } else {
        //bond 1 is directed outwards and bond 2 is directed inwards
        bond1.firstUpT = intersection(bond1FirstUp, bond1SecondUp, bond2FirstUp, bond2SecondUp);
        bond2.secondUpT = intersection(bond2FirstUp, bond2SecondUp, bond1FirstUp, bond1SecondUp);
      }
    } else {
      if (bond2IsDirectedOutwards) {
        //bond 1 is directed inwards and bond 2 is directed outwards
        bond1.secondDownT = intersection(bond1FirstDown, bond1SecondDown, bond2FirstDown, bond2SecondDown);
        bond2.firstDownT = intersection(bond2FirstDown, bond2SecondDown, bond1FirstDown, bond1SecondDown);
      } else {
        //bond 1 and bond 2 are directed inwards
        bond1.secondDownT = intersection(bond1FirstDown, bond1SecondDown, bond2FirstUp, bond2SecondUp);
        bond2.secondUpT = intersection(bond2FirstUp, bond2SecondUp, bond1FirstDown, bond1SecondDown);
      }
    }
  }
}

function getCanvas(node, canvases) {
  var canvasCandidates = canvases.filter(function (c) {
    return c.label === node.label && c.focusCharacterRange[0] === node.focusCharacterRange[0] && c.focusCharacterRange[1] === node.focusCharacterRange[1] && c.focusLine === node.focusLine;
  });
  if (canvasCandidates.length) {
    return canvasCandidates[0];
  }
  return null;
}

function getPathOfTheObjectBeingPointedAt(mousePos, objects) {
  var pathOfTheObjectBeingHoveredOver = undefined;
  objects.fragments.forEach(function (fragment, fragmentIndex) {

    fragment.bonds.forEach(function (bond, bondIndex) {
      var firstNode = fragment.nodes[bond.firstNodeIndex];
      var secondNode = fragment.nodes[bond.secondNodeIndex];
      var bondLength = Math.sqrt(Math.pow(secondNode.x - firstNode.x, 2) + Math.pow(secondNode.y - firstNode.y, 2));
      if (Math.sqrt(Math.pow(mousePos.x - firstNode.x, 2) + Math.pow(mousePos.y - firstNode.y, 2)) + Math.sqrt(Math.pow(mousePos.x - secondNode.x, 2) + Math.pow(mousePos.y - secondNode.y, 2)) < bondLength + 5) {
        pathOfTheObjectBeingHoveredOver = ['fragments', fragmentIndex, 'bonds', bondIndex];
      }
    });

    fragment.nodes.forEach(function (node, nodeIndex) {
      if (Math.sqrt(Math.pow(mousePos.x - node.x, 2) + Math.pow(mousePos.y - node.y, 2)) < 5) {
        pathOfTheObjectBeingHoveredOver = ['fragments', fragmentIndex, 'nodes', nodeIndex];
      }
    });
  });
  return pathOfTheObjectBeingHoveredOver;
}

function arraysAreEqual(a1, a2) {
  if (a1.length === a2.length) {
    for (var i = 0; i < a1.length; ++i) {
      if (a1[i] !== a2[i]) return false;
    }
    return true;
  }
  return false;
}

function arrayOfArraysIncludesArray(arrayOfArays, a) {
  return arrayOfArays.some(function (b) {
    return arraysAreEqual(b, a);
  });
}

function getHaloFromObjectPath(objectPath, objectPathIndex, hoverIsTrueSelectionIsFalse, state) {
  // Assuming objects are either nodes or bonds
  // which also means that objectPath is an array of length 4
  switch (objectPath[0]) {
    case 'fragments':
      switch (objectPath[2]) {
        case 'nodes':
          var node = state.objects.fragments[objectPath[1]].nodes[objectPath[3]];
          return _react2.default.createElement('rect', {
            key: objectPathIndex,
            x: node.x - 10,
            y: node.y - 10,
            width: 20,
            height: 20,
            style: {
              fill: hoverIsTrueSelectionIsFalse ? 'rgba(0, 0, 255, 0.5)' : 'rgb(255, 100, 255)'
            }
          });
          break;
        case 'bonds':
          var bond = state.objects.fragments[objectPath[1]].bonds[objectPath[3]];
          var firstNode = state.objects.fragments[objectPath[1]].nodes[bond.firstNodeIndex];
          var secondNode = state.objects.fragments[objectPath[1]].nodes[bond.secondNodeIndex];
          return _react2.default.createElement('line', {
            key: objectPathIndex,
            x1: firstNode.x,
            y1: firstNode.y,
            x2: secondNode.x,
            y2: secondNode.y,
            style: {
              strokeWidth: '15',
              stroke: hoverIsTrueSelectionIsFalse ? 'rgba(0, 0, 255, 0.5)' : 'rgb(100, 255, 100)',
              strokeLinecap: 'round'
            }
          });
          break;
        default:
          throw 'hmmmm';
      }
      break;
    default:
      throw 'hmmmmmm';
  }
}