'use strict';

exports.__esModule = true;

exports.default = function (fragment) {
  fragment.nodes.forEach(function (node, nodeIndex) {
    if (node.labelMode === 'CHEMICAL_ELEMENT_PLUS_FLOATING_HYDROGENS') {

      if (node.labelModeData.atomicNumber === 6) {
        // There is no label.

        // Editing the node.
        node.label = '';
        node.focusCharacterRange = [0, 0];
        node.focusLine = 0;
      } else {
        // There is a label.

        var hydrogensLabel = node.labelModeData.numberOfHydrogens ? 'H' + (node.labelModeData.numberOfHydrogens === 1 ? '' : node.labelModeData.numberOfHydrogens) : '';
        if (hydrogensLabel) {
          // We need to position the hydrogens.

          var leftBusiness = 0;
          var rightBusiness = 0;

          var bondsInvolved = (0, _helpers.getBondsInvolved)(nodeIndex, fragment.bonds);

          bondsInvolved.forEach(function (bond) {
            var bondIsDirectedOutwards = bond.firstNodeIndex === nodeIndex;

            var firstNode = fragment.nodes[bond.firstNodeIndex];
            var secondNode = fragment.nodes[bond.secondNodeIndex];

            var dX = secondNode.x - firstNode.x;
            var dY = secondNode.y - firstNode.y;

            var angle = bondIsDirectedOutwards ? Math.atan2(dY, dX) : Math.atan2(-dY, -dX);

            if (Math.abs(Math.abs(angle) - Math.PI / 2) < 0.01) {
              leftBusiness++;
              rightBusiness++;
            } else {
              if (Math.abs(angle) < Math.PI / 2) {
                rightBusiness++;
              } else {
                leftBusiness++;
              }
            }
          });

          // Editing the node.
          if (leftBusiness < rightBusiness) {
            node.label = '' + hydrogensLabel + _elements.ELEMENTS[node.labelModeData.atomicNumber].symbol;
            node.focusCharacterRange = [hydrogensLabel.length, hydrogensLabel.length + _elements.ELEMENTS[node.labelModeData.atomicNumber].symbol.length];
            node.focusLine = 0;
          } else {
            node.label = '' + _elements.ELEMENTS[node.labelModeData.atomicNumber].symbol + hydrogensLabel;
            node.focusCharacterRange = [0, _elements.ELEMENTS[node.labelModeData.atomicNumber].symbol.length];
            node.focusLine = 0;
          }
        } else {
          // There are no hydrogens.

          //Editing the node.
          node.label = _elements.ELEMENTS[node.labelModeData.atomicNumber].symbol;
          node.focusCharacterRange = [0, _elements.ELEMENTS[node.labelModeData.atomicNumber].symbol.length];
          node.focusLine = 0;
        }
      }
    }
  });
};

var _helpers = require('./helpers');

var _elements = require('./elements');