import React, { Component } from 'react';
import readMol from './readMol';
import readCdxml from './readCdxml';
import { getMousePos } from './mouseEvents';
import Fragment from './Fragment';
import Label from './Label';
import positionFloatingHydrogens from './positionFloatingHydrogens';
import create2MissingCanvases from './create2MissingCanvases';

import { getCanvas, cutBondsWithLabel, cutBondsWithoutLabel, getPathOfTheObjectBeingPointedAt, arrayOfArraysIncludesArray, getHaloFromObjectPath } from './helpers';

import CONSTANTS from './constants';

// An object can be anything: a node, a bond, or even an entire fragment!

// An object path goes like this:
// ['fragments', 3, 'nodes', 5]


function _pointerDownHandler(event) {

  const mousePos = getMousePos(this.refs.svg, event);
  const shiftKeyIsBeingPressed = event.shiftKey;

  this.setState(state => {

    mousePos.x += state.minX;
    mousePos.y += state.minY;

    const stateAssigns = {
      latestMouseDownPoint: mousePos,
      mouseIsDown: true,
      objects: {
        ...state.objects,
        fragments: state.objects.fragments.map(fragment => {
          return {
            ...fragment,
            nodes: fragment.nodes.map(node => {
              return {
                ...node,
                latestMouseDownXY: {
                  x: node.x,
                  y: node.y
                }
              };
            })
          };
        })
      }
    };

    if (shiftKeyIsBeingPressed) {
      const pathOfTheObjectBeingPointedAt = getPathOfTheObjectBeingPointedAt(mousePos, state.objects);

      if (pathOfTheObjectBeingPointedAt) {

        if (arrayOfArraysIncludesArray(state.pathsOfSelectedObjects, pathOfTheObjectBeingPointedAt)) {
          // We are hitting one of the selected objects
          return {
            ...state,
            ...stateAssigns
          };
        }

        // We are hitting something unselected.
        // What we are hitting is selected in addition to the objects that have already been slected (if any)
        // because the Shift key is being pressed.
        return {
          ...state,
          pathsOfSelectedObjects: [
            ...state.pathsOfSelectedObjects,
            pathOfTheObjectBeingPointedAt
          ],
          ...stateAssigns
        };
      }
    } else {
      const pathOfTheObjectBeingPointedAt = getPathOfTheObjectBeingPointedAt(mousePos, state.objects);

      if (pathOfTheObjectBeingPointedAt) {

        if (arrayOfArraysIncludesArray(state.pathsOfSelectedObjects, pathOfTheObjectBeingPointedAt)) {
          // We are hitting one of the selected objects
          return {
            ...state,
            ...stateAssigns
          };
        }

        // We are hitting something unselected.
        // If any objects are selected, they are deselected
        // because the Shift key is not pressed.
        // What we are hitting is selected.
        return {
          ...state,
          pathsOfSelectedObjects: [
            pathOfTheObjectBeingPointedAt
          ],
          ...stateAssigns
        };
      }

      // We are hitting nowhere.
      return {
        ...state,
        pathsOfSelectedObjects: [],
        ...stateAssigns
      };
    }

    return {
      ...state,
      ...stateAssigns
    };

  });
}


function _pointerUpHandler(event) {

  const mousePos = getMousePos(this.refs.svg, event);

  this.setState(state => {

    mousePos.x += state.minX;
    mousePos.y += state.minY;

    // let hitFragmentIndex = undefined;
    // let hitNodeIndex = undefined;
    //
    // state.objects.fragments.forEach((fragment, fragmentIndex) => {
    //   fragment.nodes.forEach((node, nodeIndex) => {
    //     if (Math.abs(node.x - mousePos.x) < 20 && Math.abs(node.y - mousePos.y) < 20) {
    //       hitFragmentIndex = fragmentIndex;
    //       hitNodeIndex = nodeIndex;
    //     }
    //   });
    // });



    if (state.latestMouseDownPoint.x === mousePos.x && state.latestMouseDownPoint.y === mousePos.y) {
      // We clicked and released the pointer button.

      // const pathOfTheObjectHavingBeenClicked = getPathOfTheObjectBeingPointedAt(mousePos, state.objects);
      //
      // if (pathOfTheObjectHavingBeenClicked) {
      //   return {
      //     ...state,
      //     pathsOfSelectedObjects: [
      //       pathOfTheObjectHavingBeenClicked
      //     ],
      //     mouseIsDown: false
      //   };
      // }
      //
      // // We clicked nowhere.
      // return {
      //   ...state,
      //   pathsOfSelectedObjects: [],
      //   mouseIsDown: false
      // };


      // return {
      //   ...state,
      //   objects: {
      //     ...state.objects,
      //     fragments: state.objects.fragments.map((fragment, fragmentIndex) => {
      //       return {
      //         ...fragment,
      //         nodes: fragment.nodes.map((node, nodeIndex) => {
      //           return {
      //             ...node,
      //             beingEdited: (fragmentIndex === hitFragmentIndex && nodeIndex === hitNodeIndex)
      //           };
      //         })
      //       };
      //     })
      //   }
      // };

      // if (typeof hitFragmentIndex !== 'undefined' && typeof hitNodeIndex !== 'undefined') {
      //   // we clicked on a node
      // }

    }

    // No changes.
    return {
      ...state,
      mouseIsDown: false
    };;
  });
}


function _pointerMoveHandler(event) {

  const state = this.state;
  if (state.mouseIsDown && state.pathsOfSelectedObjects.length) {
    // Mouse is pressed and something is selected, so we are dragging
    // the selected objects.
    event.preventDefault();
  }

  const mousePos = getMousePos(this.refs.svg, event);
  this.setState(state => {

    mousePos.x += state.minX;
    mousePos.y += state.minY;

    if (state.mouseIsDown) {
      if (state.pathsOfSelectedObjects.length) {
        // Mouse is pressed and something is selected, so we are dragging
        // the selected objects.

        let newCanvasCache = [...state.canvasCache];

        const newObjects = {
          ...state.objects,
          fragments: state.objects.fragments.map((fragment, fragmentIndex) => {
            const newFragment = {
              ...fragment,
              nodes: fragment.nodes.map((node, nodeIndex) => {
                if (arrayOfArraysIncludesArray(state.pathsOfSelectedObjects, ['fragments', fragmentIndex, 'nodes', nodeIndex])) {
                  return {
                    ...node,
                    x: mousePos.x - state.latestMouseDownPoint.x + node.latestMouseDownXY.x,
                    y: mousePos.y - state.latestMouseDownPoint.y + node.latestMouseDownXY.y
                  };
                }
                return { ...node };
              }),
              bonds: fragment.bonds.map(bond => ({ ...bond }))
            };

            positionFloatingHydrogens(newFragment);

            const newCs = create2MissingCanvases(newFragment.nodes, newCanvasCache);
            newCanvasCache = [...newCanvasCache, ...newCs];

            // DETERMINING BOND POLYGONS IN THE new fragment !!!!! new !!!!!
            newFragment.nodes.forEach((node, nodeIndex) => {
              const c = getCanvas(node, newCanvasCache);
              if (!c) throw 'hmmmm';
              if (node.label) {
                cutBondsWithLabel(nodeIndex, c.filledPixelsData, newFragment.nodes, newFragment.bonds);
              } else {
                cutBondsWithoutLabel(nodeIndex, newFragment.nodes, newFragment.bonds);
              }
            });
            // DETERMINING BOND POLYGONS IN THE new fragment !!!!! new !!!!!

            return newFragment;
          })
        };

        // Because we have moved nodes, we need to determine hovered
        // objects using newObjects
        return {
          ...state,
          objects: newObjects,
          pathOfTheObjectBeingHoveredOver: getPathOfTheObjectBeingPointedAt(mousePos, newObjects),
          canvasCache: newCanvasCache
        };
      }

    }

    return {
      ...state,
      pathOfTheObjectBeingHoveredOver: getPathOfTheObjectBeingPointedAt(mousePos, state.objects)
    };
  });
  //
  // const mousePos = getMousePos(this.refs.svg, event);
  // console.log(mousePos);
  // console.log(this.state);
  //
  // this.setState(state => {
  //   return {
  //     ...state,
  //     // cursor: (Math.random() < 0.5 ? 'auto' : 'pointer'),
  //     objects: {
  //       ...state.objects,
  //       fragments: state.objects.fragments.map(fragment => ({
  //         ...fragment,
  //         nodes: fragment.nodes.map(node => ({
  //           ...node,
  //           hover: (node.x <= mousePos.x && mousePos.x < node.x + 6 && node.y <= mousePos.y && mousePos.y < node.y + 6)
  //         }))
  //       }))
  //     }
  //   };
  // })
}


class Canvas extends Component {

  constructor(props) {
    super(props);
    this.state = {
      mode: 'VIEWING', // VIEWING, EDITING_LABEL, DRAGGING(?)
      modeData: {
      },
      // For VIEWING we have an empty modeData object.
      // For EDITING_LABEL we have pathOfTheObjectBeingEdited.
      // For DRAGGING we have an empty modeData object.

      // We render selection & hover halos underneath/below everything else
      pathsOfSelectedObjects: [
        // ['fragments', 1, 'bonds', 4]
      ],
      pathOfTheObjectBeingHoveredOver: undefined,


      latestMouseDownPoint: { // start drag point
        x: undefined,
        y: undefined
      },

      mouseIsDown: false,
      // cursor: 'auto',
      // toolSelected: 'SELECT',
      objects: {
        fragments: [
          {
            nodes: [],
            bonds: []
          }
        ]
      },
      canvasCache: [
        // {
        //   label: 'C',
        //   focusCharacterRange: [0, atomLine.substring(31,34).trim().length], // one range per line
        //   focusLine: 0, // assuming one line
        //   filledPixelsData: Float32Array(flatten([[0, 0], [0, 1], ...])),
        //   labelOriginXRelativeToFocus: -5.345,
        //   labelOriginYRelativeToFocus: 1.435
        // }
      ],
      minX: 99999, // We will apply Math.min to it.
      minY: 99999, // We will apply Math.min to it.
      maxX: 0, // We will apply Math.max to it.
      maxY: 0 // We will apply Math.max to it.
    };

    this.pointerUpHandler = _pointerUpHandler.bind(this);
    this.pointerDownHandler = _pointerDownHandler.bind(this);
    this.pointerMoveHandler = _pointerMoveHandler.bind(this);
  }

  componentWillUnmount() {
    this.refs.svg.removeEventListener('touchmove', this.pointerMoveHandler, { passive: false }, false);
  }

  componentDidMount() {
    this.refs.svg.addEventListener('touchmove', this.pointerMoveHandler, { passive: false }, false);


    // this.placeMol(
    //   `\n\n\n 25 26  0  0001  0  0  0  0  0999 V2000\n    9.4638   -0.9375    0.0000 S   0  0  0  0  0  0  0  0  0  0  0\n   12.6690   -4.3708    0.0000 Na  0  3  0  0  0  0  0  0  0  0  0\n    5.7264   -3.3446    0.0000 O   0  0  0  0  0  0  0  0  0  0  0\n   11.2628   -4.3455    0.0000 O   0  5  0  0  0  0  0  0  0  0  0\n    9.1344   -5.4097    0.0000 O   0  0  0  0  0  0  0  0  0  0  0\n    4.5989   -2.6732    0.0000 O   0  0  0  0  0  0  0  0  0  0  0\n    8.1969   -2.6732    0.0000 N   0  0  0  0  0  0  0  0  0  0  0\n    5.7264   -0.6715    0.0000 N   0  0  0  0  0  0  0  0  0  0  0\n    8.1969   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    6.8666   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    9.4638   -3.0786    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    6.8666   -2.6732    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n   10.2619   -2.0017    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    9.9325   -4.3455    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n   11.1994   -2.9392    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n   11.1994   -1.0769    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    4.5989   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    3.4333   -0.6715    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    2.3058   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    1.1529   -0.6715    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    2.3058   -2.6732    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    0.0000   -1.3429    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    1.1529   -3.3320    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    0.0000   -2.6732    0.0000 C   0  0  0  0  0  0  0  0  0  0  0\n    8.1969    0.0000    0.0000 H   0  0  0  0  0  0  0  0  0  0  0\n  1  9  1  0\n  1 13  1  0\n  3 12  2  0\n  4 14  1  0\n  5 14  2  0\n  6 17  2  0\n  7  9  1  0\n  7 11  1  0\n  7 12  1  0\n 10  8  1  1\n  8 17  1  0\n  9 10  1  0\n  9 25  1  6\n 10 12  1  0\n 11 13  1  0\n 11 14  1  6\n 13 15  1  0\n 13 16  1  0\n 17 18  1  0\n 18 19  1  0\n 19 20  2  0\n 19 21  1  0\n 20 22  1  0\n 21 23  2  0\n 22 24  2  0\n 23 24  1  0\nM  CHG  2   2   1   4  -1\nM  END\n`,
    //   this.state
    // );
    this.placeCdxml(`<?xml version="1.0" encoding="UTF-8" ?>
<!DOCTYPE CDXML SYSTEM "http://www.cambridgesoft.com/xml/cdxml.dtd" >
<CDXML
 CreationProgram="ChemDraw 14.0.0.117"
 Name="Untitled ACS Document 1996-4"
 BoundingBox="118 78.45 239.82 187.74"
 WindowPosition="0 0"
 WindowSize="0 -2147483648"
 WindowIsZoomed="yes"
 FractionalWidths="yes"
 InterpretChemically="yes"
 ShowAtomQuery="yes"
 ShowAtomStereo="no"
 ShowAtomEnhancedStereo="yes"
 ShowAtomNumber="no"
 ShowResidueID="no"
 ShowBondQuery="yes"
 ShowBondRxn="yes"
 ShowBondStereo="no"
 ShowTerminalCarbonLabels="no"
 ShowNonTerminalCarbonLabels="no"
 HideImplicitHydrogens="no"
 LabelFont="90"
 LabelSize="10"
 LabelFace="96"
 CaptionFont="4"
 CaptionSize="12"
 HashSpacing="2.70"
 MarginWidth="2"
 LineWidth="1"
 BoldWidth="4"
 BondLength="28.35"
 BondSpacing="12"
 ChainAngle="120"
 LabelJustification="Auto"
 CaptionJustification="Left"
 AminoAcidTermini="HOH"
 ShowSequenceTermini="yes"
 ShowSequenceBonds="yes"
 ResidueWrapCount="40"
 ResidueBlockCount="10"
 ResidueZigZag="yes"
 NumberResidueBlocks="no"
 PrintMargins="36 36 36 36"
 MacPrintInfo="0003000001200120000000000B6608A0FF84FF880BE309180367052703FC0002000001200120000000000B6608A0000100640064000000010001010100000001270F000100010000000000000000000000000002001901900000000000600000000000000000000100000000000000000000000000000000"
 ChemPropName=""
 ChemPropFormula="Chemical Formula: "
 ChemPropExactMass="Exact Mass: "
 ChemPropMolWt="Molecular Weight: "
 ChemPropMOverZ="m/z: "
 ChemPropAnalysis="Elemental Analysis: "
 ChemPropBoilingPt="Boiling Point: "
 ChemPropMeltingPt="Melting Point: "
 ChemPropCritTemp="Critical Temp: "
 ChemPropCritPres="Critical Pres: "
 ChemPropCritVol="Critical Vol: "
 ChemPropGibbs="Gibbs Energy: "
 ChemPropLogP="Log P: "
 ChemPropMR="MR: "
 ChemPropHenry="Henry&apos;s Law: "
 ChemPropEForm="Heat of Form: "
 ChemProptPSA="tPSA: "
 ChemPropCLogP="CLogP: "
 ChemPropCMR="CMR: "
 ChemPropLogS="LogS: "
 ChemPropPKa="pKa: "
 color="0"
 bgcolor="1"
><colortable>
<color r="1" g="1" b="1"/>
<color r="0" g="0" b="0"/>
<color r="1" g="0" b="0"/>
<color r="1" g="1" b="0"/>
<color r="0" g="1" b="0"/>
<color r="0" g="1" b="1"/>
<color r="0" g="0" b="1"/>
<color r="1" g="0" b="1"/>
</colortable><fonttable>
<font id="3" charset="iso-8859-1" name="Arial"/>
<font id="4" charset="iso-8859-1" name="Times New Roman"/>
<font id="90" charset="iso-8859-1" name="Helvetica"/>
</fonttable><page
 id="1472"
 BoundingBox="0 0 540 719.75"
 HeaderPosition="36"
 FooterPosition="36"
 PrintTrimMarks="yes"
 HeightPages="1"
 WidthPages="1"
><fragment
 id="1435"
 BoundingBox="118 78.45 239.82 187.74"
 Z="26"
><n
 id="1432"
 p="170.05 132.03"
 Z="27"
 Element="7"
 NumHydrogens="1"
 NeedsClean="yes"
 AS="N"
><t
 p="166.65 135.71"
 BoundingBox="166.65 127.83 180.25 135.71"
 LabelJustification="Left"
 LabelAlignment="Left"
><s font="3" size="9.4" face="96">NH</s></t></n><n
 id="1434"
 p="170.05 103.68"
 Z="28"
 Element="7"
 NumHydrogens="1"
 NeedsClean="yes"
 AS="N"
><t
 p="166.65 107.36"
 BoundingBox="166.65 99.48 180.25 107.36"
 LabelJustification="Left"
 LabelAlignment="Left"
><s font="3" size="9.4" face="96">NH</s></t></n><n
 id="1436"
 p="141.70 103.68"
 Z="29"
/><n
 id="1438"
 p="141.70 132.03"
 Z="30"
 AS="N"
/><n
 id="1444"
 p="194.60 117.86"
 Z="34"
 Element="7"
 NumHydrogens="1"
 NeedsClean="yes"
 AS="N"
><t
 p="198 121.53"
 BoundingBox="184.40 113.66 198 121.53"
 LabelJustification="Right"
 Justification="Right"
 LabelAlignment="Right"
><s font="3" size="9.4" face="96">NH</s></t></n><n
 id="1446"
 p="208.78 93.31"
 Z="35"
/><n
 id="1448"
 p="184.22 79.13"
 Z="36"
 AS="N"
/><n
 id="1454"
 p="184.22 156.59"
 Z="40"
/><n
 id="1456"
 p="208.78 142.41"
 Z="41"
 AS="N"
/><n
 id="1462"
 p="121.65 83.64"
 Z="45"
 Element="8"
 NumHydrogens="0"
 NeedsClean="yes"
 AS="N"
><t
 p="118 87.31"
 BoundingBox="118 79.32 125.31 87.41"
 LabelJustification="Left"
><s font="3" size="9.4" face="96">O</s></t></n><n
 id="1465"
 p="176.89 183.97"
 Z="47"
 Element="8"
 NumHydrogens="0"
 NeedsClean="yes"
 AS="N"
><t
 p="173.24 187.64"
 BoundingBox="173.24 179.65 180.55 187.74"
 LabelJustification="Left"
><s font="3" size="9.4" face="96">O</s></t></n><n
 id="1468"
 p="236.16 85.97"
 Z="49"
 Element="8"
 NumHydrogens="0"
 NeedsClean="yes"
 AS="N"
><t
 p="232.51 89.64"
 BoundingBox="232.51 81.65 239.82 89.74"
 LabelJustification="Left"
><s font="3" size="9.4" face="96">O</s></t></n><b
 id="1441"
 Z="31"
 B="1434"
 E="1436"
 BS="N"
/><b
 id="1442"
 Z="32"
 B="1436"
 E="1438"
 BS="N"
/><b
 id="1443"
 Z="33"
 B="1438"
 E="1432"
 BS="N"
/><b
 id="1451"
 Z="37"
 B="1444"
 E="1446"
 BS="N"
/><b
 id="1452"
 Z="38"
 B="1446"
 E="1448"
 BS="N"
/><b
 id="1453"
 Z="39"
 B="1448"
 E="1434"
 BS="N"
/><b
 id="1459"
 Z="42"
 B="1432"
 E="1454"
 BS="N"
/><b
 id="1460"
 Z="43"
 B="1454"
 E="1456"
 BS="N"
/><b
 id="1461"
 Z="44"
 B="1456"
 E="1444"
 BS="N"
/><b
 id="1464"
 Z="46"
 B="1436"
 E="1462"
 Order="2"
/><b
 id="1467"
 Z="48"
 B="1454"
 E="1465"
 Order="2"
/><b
 id="1470"
 Z="50"
 B="1446"
 E="1468"
 Order="2"
/></fragment></page></CDXML>`, this.state);
    // this.setState(state => {
    //   return {
    //     ...state,
    //     pathsOfSelectedObjects: [
    //       ['fragments', 1, 'bonds', 4],
    //       ['fragments', 1, 'bonds', 3]
    //     ]
    //   };
    // });
  }

  placeMol(mol, state) {


    const fragment = readMol(mol);

    const newCs = create2MissingCanvases(fragment.nodes, state.canvasCache);


    // DETERMINING BOND POLYGONS IN THE fragment
    fragment.nodes.forEach((node, nodeIndex) => {
      const c = getCanvas(node, [...state.canvasCache, ...newCs]);
      if (!c) throw 'hmmmm';
      if (node.label) {
        cutBondsWithLabel(nodeIndex, c.filledPixelsData, fragment.nodes, fragment.bonds);
      } else {
        cutBondsWithoutLabel(nodeIndex, fragment.nodes, fragment.bonds);
      }
    });
    // DETERMINING BOND POLYGONS IN THE fragment



    let newMinX = state.minX;
    let newMinY = state.minY;
    let newMaxX = state.maxX;
    let newMaxY = state.maxY;
    fragment.nodes.forEach(node => {
      const c = getCanvas(node, [...state.canvasCache, ...newCs]);
      if (!c) throw 'hmmmm';
      newMinX = Math.min(newMinX, node.x + c.minX - CONSTANTS.SVG_PADDING);
      newMinY = Math.min(newMinY, node.y + c.minY - CONSTANTS.SVG_PADDING);
      newMaxX = Math.max(newMaxX, node.x + c.maxX + CONSTANTS.SVG_PADDING);
      newMaxY = Math.max(newMaxY, node.y + c.maxY + CONSTANTS.SVG_PADDING);
    });

    this.setState(state => {
      return {
        ...state,
        objects: {
          ...state.objects,
          fragments: [
            ...state.objects.fragments,
            fragment
          ]
        },
        canvasCache: [...state.canvasCache, ...newCs],
        minX: newMinX,
        minY: newMinY,
        maxX: newMaxX,
        maxY: newMaxY
      };
    });
  }

  placeCdxml(cdxml, state) {


    const fragment = readCdxml(cdxml);

    const newCs = create2MissingCanvases(fragment.nodes, state.canvasCache);


    // DETERMINING BOND POLYGONS IN THE fragment
    fragment.nodes.forEach((node, nodeIndex) => {
      const c = getCanvas(node, [...state.canvasCache, ...newCs]);
      if (!c) throw 'hmmmm';
      if (node.label) {
        cutBondsWithLabel(nodeIndex, c.filledPixelsData, fragment.nodes, fragment.bonds);
      } else {
        cutBondsWithoutLabel(nodeIndex, fragment.nodes, fragment.bonds);
      }
    });
    // DETERMINING BOND POLYGONS IN THE fragment



    let newMinX = state.minX;
    let newMinY = state.minY;
    let newMaxX = state.maxX;
    let newMaxY = state.maxY;
    fragment.nodes.forEach(node => {
      const c = getCanvas(node, [...state.canvasCache, ...newCs]);
      if (!c) throw 'hmmmm';
      newMinX = Math.min(newMinX, node.x + c.minX - CONSTANTS.SVG_PADDING);
      newMinY = Math.min(newMinY, node.y + c.minY - CONSTANTS.SVG_PADDING);
      newMaxX = Math.max(newMaxX, node.x + c.maxX + CONSTANTS.SVG_PADDING);
      newMaxY = Math.max(newMaxY, node.y + c.maxY + CONSTANTS.SVG_PADDING);
    });

    this.setState(state => {
      return {
        ...state,
        objects: {
          ...state.objects,
          fragments: [
            ...state.objects.fragments,
            fragment
          ]
        },
        canvasCache: [...state.canvasCache, ...newCs],
        minX: newMinX,
        minY: newMinY,
        maxX: newMaxX,
        maxY: newMaxY
      };
    });
  }



  setLabel(fragmentIndex, nodeIndex, newLabel, state) {

    const oldFragment = state.objects.fragments[fragmentIndex];
    const newNodes = [...oldFragment.nodes];
    newNodes[nodeIndex] = {
      ...newNodes[nodeIndex],
      label: newLabel
    };
    const newFragment = {
      ...oldFragment,
      nodes: newNodes
    };


    const newCs = create2MissingCanvases(newFragment.nodes, state.canvasCache);


    // DETERMINING BOND POLYGONS IN THE newFragment !!!! new !!!!!
    newFragment.nodes.forEach((node, nodeIndex) => {
      const c = getCanvas(node, [...state.canvasCache, ...newCs]);
      if (!c) throw 'hmmmm';
      if (node.label) {
        cutBondsWithLabel(nodeIndex, c.filledPixelsData, newFragment.nodes, newFragment.bonds);
      } else {
        cutBondsWithoutLabel(nodeIndex, newFragment.nodes, newFragment.bonds);
      }
    });
    // DETERMINING BOND POLYGONS IN THE newFragment !!!! new !!!!!






    const newFragments = state.objects.fragments;
    newFragments[fragmentIndex] = newFragment;
    this.setState(state => {
      return {
        ...state,
        objects: {
          ...state.objects,
          fragments: newFragments
        },
        canvasCache: [...state.canvasCache, ...newCs]
      };
    });

  }

  renderSelectedObjectHalos(state) {

    // Returning an array of two elements: selection halos and hover halo
    return [
      state.pathsOfSelectedObjects.map((objectPath, objectPathIndex) => getHaloFromObjectPath(objectPath, objectPathIndex, false, state)),
      state.pathOfTheObjectBeingHoveredOver && getHaloFromObjectPath(state.pathOfTheObjectBeingHoveredOver, state.pathsOfSelectedObjects.length, true, state)
    ];
  }



  render() {
    const state = this.state;

    return (
      <div style={{
        position: 'relative',
        width: `${state.maxX - state.minX}px`,
        height: `${state.maxY - state.minY}px`,
        display: 'inline-block'
        //opacity: 0.05
      }}>
        <svg
          viewBox={`${state.minX} ${state.minY} ${state.maxX - state.minX} ${state.maxY - state.minY}`}
          style={{
            position: 'absolute',
            top: '0',
            left: '0',
            cursor: state.pathOfTheObjectBeingHoveredOver ? 'pointer' : 'auto',
            width: `${state.maxX - state.minX}px`,
            height: `${state.maxY - state.minY}px`
          }}
          ref="svg"
          width={state.maxX - state.minX}
          height={state.maxY - state.minY}
          onMouseDown={this.pointerDownHandler}
          onMouseUp={this.pointerUpHandler}
          onMouseMove={this.pointerMoveHandler}
          onTouchStart={this.pointerDownHandler}
          onTouchEnd={this.pointerUpHandler}
        >
          {this.renderSelectedObjectHalos(state)}
          {state.objects.fragments.map((fragment, fragmentIndex) => (
            <Fragment
              key={fragmentIndex}
              canvasCache={state.canvasCache}
              fragment={fragment}
            />
          ))}
        </svg>
        {state.objects.fragments.map((fragment, fragmentIndex) => {
          return <div key={fragmentIndex} style={{ position: 'absolute', top: '0', left: '0' }}>
            {fragment.nodes.map((node, nodeIndex) => {
              if (!node.beingEdited) return;
              const c = getCanvas(node, state.canvasCache);
              if (!c) return;
              return (
                <Label
                  key={nodeIndex}
                  origin={{
                    x: node.x + c.labelOriginXRelativeToFocus,
                    y: node.y + c.labelOriginYRelativeToFocus
                  }}
                  onChange={newLabel => {
                    this.setLabel(fragmentIndex, nodeIndex, newLabel, state);
                  }}
                  value={node.label}
                />
              )
            })}
          </div>
        })}
      </div>
    );
  }
}

export default Canvas;
