import React, { Component } from 'react';
import PropTypes from 'prop-types';
import CONSTANTS from './constants';

// We use contentEditable span and not input.
// Benefits of input:
// (a) no formatting,
// (b) always single line,
// (c) easier onChange handlers?
// Benefits of contentEditable:
// (a) stretches on demand,
// (b) we can house formatted content, i.e. it's future proofed

// Need to copy the way draft js handles onpaste (we don't want to paste
// formatted content), entities (our focus range is an entity)
// Use mutation observer (instead of onInput)?
class Label extends Component {

  componentDidMount() {
    this.refs.span.textContent = this.props.value;
  }

  // componentDidUpdate(prevProps, prevState) {
  //
  //   const div = document.createElement('DIV');
  //   div.textContent = this.props.value;
  //
  //   if (this.refs.span.innerHTML !== div.innerHTML) {
  //     this.refs.span.innerHTML = div.innerHTML;
  //   }
  // }


  render() {
    return (
      <div style={{
        position: 'absolute',
        left: `${this.props.origin.x}px`,
        top: `${this.props.origin.y - 5*CONSTANTS.FONT_SIZE}px`
      }}>
        <span
          ref="span"
          style={{
            display: 'inline-block',
            padding: '10px 0',

            fontSize: `${CONSTANTS.FONT_SIZE}px`,

            opacity: '0.75',

            border: 'none',
            outline: 'none',
            appearance: 'none',
            WebkitAppearance: 'none',
            backgroundColor: 'transparent',
            font: `${CONSTANTS.FONT_SIZE}px sans-serif`,

            lineHeight: `${CONSTANTS.FONT_SIZE}px`,
            whiteSpace: 'nowrap'
          }}
          onInput={() => {
            this.props.onChange(this.refs.span.textContent);
          }}
          // onKeyDown={() => {
          //   setTimeout(() => {
          //     this.props.onChange(this.refs.span.textContent);
          //   }, 1);
          // }}
          contentEditable={true}
        />
        <span style={{
          display: 'inline-block',
          height: `${5*CONSTANTS.FONT_SIZE}px`
        }} />
      </div>
    );
  }

}

Label.propTypes = {
  origin: PropTypes.object.isRequired, // intersection of left edge and baseline
  onChange: PropTypes.func.isRequired, // accepts one parameter which is new label
  value: PropTypes.string.isRequired
}

export default Label;
