import { subtractPoints, createPoint, slidePoint, addPoints, scalePoint, pointToPathString } from './helpers';
import CONSTANTS from './constants';

export function getSingleBondPath(bond, firstNode, secondNode) {
  const d = subtractPoints(secondNode, firstNode);
  const length = Math.sqrt(d.x * d.x + d.y * d.y);
  const bondVerticalUnitVector = createPoint(0 - d.y/length, d.x/length);

  let firstEndPath;
  if (firstNode.label) {
    firstEndPath = `${
      pointToPathString(slidePoint(
       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
       bond.firstT
     ))
    } L ${
      pointToPathString(slidePoint(
        addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
        addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
        bond.firstT
      ))
    }`;
  } else {
    firstEndPath = `${
      pointToPathString(slidePoint(
       addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
       addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
       bond.firstUpT
     ))
   } L ${
     pointToPathString(firstNode)
   } L ${
      pointToPathString(slidePoint(
        addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
        addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
        bond.firstDownT
      ))
    }`;
  }

  let secondEndPath;
  if (secondNode.label) {
    secondEndPath = `${
      pointToPathString(slidePoint(
       addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
       addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
       bond.secondT
     ))
    } L ${
      pointToPathString(slidePoint(
        addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
        addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
        bond.secondT
      ))
    }`;
  } else {
    secondEndPath = `${
      pointToPathString(slidePoint(
       addPoints(firstNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
       addPoints(secondNode, scalePoint(bondVerticalUnitVector, -CONSTANTS.BOND_BAR_WIDTH/2)),
       bond.secondDownT
     ))
    } L ${
      pointToPathString(secondNode)
    } L ${
      pointToPathString(slidePoint(
        addPoints(firstNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
        addPoints(secondNode, scalePoint(bondVerticalUnitVector, CONSTANTS.BOND_BAR_WIDTH/2)),
        bond.secondUpT
      ))
    }`;
  }

  return `M ${firstEndPath} L ${secondEndPath} Z`;
}
