import { getCanvas } from './helpers';
import CONSTANTS from './constants';

// The function returns an array of canvases that don't exist in canvasCache.
export default function(nodes, canvasCache) {
  const newCs = [];

  nodes.forEach(node => {

    const existingC = getCanvas(node, [...canvasCache, ...newCs]);
    if (existingC) return;

    const canvasDomElement = document.createElement('CANVAS');
    const ctx = canvasDomElement.getContext('2d');
    ctx.font = `${CONSTANTS.FONT_SIZE}px sans-serif`;
    ctx.strokeStyle = 'red';
    ctx.lineWidth = CONSTANTS.LABEL_BOND_SPACING;
    ctx.lineJoin = 'round';
    // const nodeLabelLines = node.label.split('\n');

    // Assuming one line. (We call it label for now.)
    // Assuming no kerning or ligatures.
    // Y is directed downwards.
    // We will have one focus character range per line; center of the range is the center of the line.
    // Origin is the point we specify when using ctx.fillText/strokeText (left bottom point of text).
    // Bottom of the label is the baseline. Height is font size.
    const labelOriginXRelativeToFocus = - ctx.measureText(node.label.substring(0, node.focusCharacterRange[0])).width - ctx.measureText(node.label.substring(node.focusCharacterRange[0], node.focusCharacterRange[1])).width / 2;
    const labelOriginYRelativeToFocus = CONSTANTS.FONT_SIZE / 3;
    const labelWidth = ctx.measureText(node.label).width;
    const labelHeight = CONSTANTS.FONT_SIZE;

    // Adding padding to our canvases for "ygp..." protrusions etc.
    const leftCanvasBorderXRelativeToFocus = labelOriginXRelativeToFocus - CONSTANTS.ADDED_MARGIN_FOR_CANVASES;
    const rightCanvasBorderXRelativeToFocus = labelOriginXRelativeToFocus + labelWidth + CONSTANTS.ADDED_MARGIN_FOR_CANVASES;
    const topCanvasBorderYRelativeToFocus = labelOriginYRelativeToFocus - labelHeight - CONSTANTS.ADDED_MARGIN_FOR_CANVASES;
    const bottomCanvasBorderYRelativeToFocus = labelOriginYRelativeToFocus + CONSTANTS.ADDED_MARGIN_FOR_CANVASES;

    // const textWidth = Math.max(...nodeLabelLines.map(labelLine => ctx.measureText(labelLine).width));
    // console.log(textWidth);
    // const textHeight = nodeLabelLines.length * CONSTANTS.FONT_SIZE;
    canvasDomElement.width = rightCanvasBorderXRelativeToFocus - leftCanvasBorderXRelativeToFocus;
    canvasDomElement.height = bottomCanvasBorderYRelativeToFocus - topCanvasBorderYRelativeToFocus;

    // Our ctx got reset after we resized the canvas.
    ctx.font = `${CONSTANTS.FONT_SIZE}px sans-serif`;
    ctx.strokeStyle = 'red';
    ctx.lineWidth = CONSTANTS.LABEL_BOND_SPACING;
    ctx.lineJoin = 'round';



    // nodeLabelLines.forEach((labelLine, labelLineIndex) => {
    //   ctx.strokeText(labelLine, ADDED_MARGIN_FOR_CANVASES, ADDED_MARGIN_FOR_CANVASES + (labelLineIndex + 1) * CONSTANTS.FONT_SIZE);
    // });
    ctx.strokeText(node.label, labelOriginXRelativeToFocus - leftCanvasBorderXRelativeToFocus, labelOriginYRelativeToFocus - topCanvasBorderYRelativeToFocus);

    const pixelData = ctx.getImageData(0, 0, canvasDomElement.width, canvasDomElement.height).data;

    const filledPixels = [];
    for (let x = 0; x < canvasDomElement.width; ++x) {
      for (let y = 0; y < canvasDomElement.height; ++y) {
        if (pixelData[y*canvasDomElement.width*4 + x*4 + 3] > 0) { // max for alpha is 255?
          filledPixels.push([x + leftCanvasBorderXRelativeToFocus + 0.5, y + topCanvasBorderYRelativeToFocus + 0.5]);
        }
      }
    }


    // The Y coordinate line is directed downwards.
    let minX = 99999; // We will apply Math.min to it.
    let minY = 99999; // We will apply Math.min to it.
    let maxX = 0; // We will apply Math.max to it.
    let maxY = 0; // // We will apply Math.max to it.

    filledPixels.forEach(filledPixel => {
      minX = Math.min(minX, filledPixel[0]);
      minY = Math.min(minY, filledPixel[1]);
      maxX = Math.max(maxX, filledPixel[0]);
      maxY = Math.max(maxY, filledPixel[1]);
    });


    const filledPixelsFlat = filledPixels.reduce((acc, filledPixel) => {
      acc.push(filledPixel[0]);
      acc.push(filledPixel[1]);
      return acc;
    }, []);
    const filledPixelsData = new Float32Array(filledPixelsFlat);

    const c = {
      label: node.label,
      focusCharacterRange: node.focusCharacterRange, // one range per line
      focusLine: node.focusLine,
      filledPixels,
      filledPixelsData, // origin here is the focus point which is in the middle of focus line (both in X and in Y coordinates)
      labelOriginXRelativeToFocus,
      labelOriginYRelativeToFocus,
      minX,
      minY,
      maxX,
      maxY
    };

    // this.setState(state => {
    //   // if (state.canvasCache.some(c => {
    //   //   return c.label === node.label;
    //   // })) {
    //   //   return state;
    //   // }
    //   return {
    //     ...state,
    //     canvasCache: [...state.canvasCache, c]
    //   };
    // });


    newCs.push(c);
    //return c;
  });
  return newCs;
}
