import { getBondsInvolved } from './helpers';
import { ELEMENTS } from './elements';

// The function mutates the fragment's nodes.
export default function(fragment) {
  fragment.nodes.forEach((node, nodeIndex) => {
    if (node.labelMode === 'CHEMICAL_ELEMENT_PLUS_FLOATING_HYDROGENS') {

      if (node.labelModeData.atomicNumber === 6) {
        // There is no label.

        // Editing the node.
        node.label = '';
        node.focusCharacterRange = [0, 0];
        node.focusLine = 0;
      } else {
        // There is a label.

        const hydrogensLabel = node.labelModeData.numberOfHydrogens ? `H${node.labelModeData.numberOfHydrogens === 1 ? '' : node.labelModeData.numberOfHydrogens}` : '';
        if (hydrogensLabel) {
          // We need to position the hydrogens.

          let leftBusiness = 0;
          let rightBusiness = 0;

          const bondsInvolved = getBondsInvolved(nodeIndex, fragment.bonds);

          bondsInvolved.forEach(bond => {
            const bondIsDirectedOutwards = (bond.firstNodeIndex === nodeIndex);

            const firstNode = fragment.nodes[bond.firstNodeIndex];
            const secondNode = fragment.nodes[bond.secondNodeIndex];

            const dX = secondNode.x - firstNode.x;
            const dY = secondNode.y - firstNode.y;

            const angle = bondIsDirectedOutwards ? Math.atan2(dY, dX) : Math.atan2(-dY, -dX);

      		  if (Math.abs(Math.abs(angle) - Math.PI/2) < 0.01) {
      			  leftBusiness++;
      			  rightBusiness++;
      		  } else {
      			  if (Math.abs(angle) < Math.PI/2) {
      				      rightBusiness++;
      			  } else {
      				    leftBusiness++;
      			  }
      		  }
          });

          // Editing the node.
          if (leftBusiness < rightBusiness) {
            node.label = `${hydrogensLabel}${ELEMENTS[node.labelModeData.atomicNumber].symbol}`;
            node.focusCharacterRange = [hydrogensLabel.length, hydrogensLabel.length + ELEMENTS[node.labelModeData.atomicNumber].symbol.length];
            node.focusLine = 0;
          } else {
            node.label = `${ELEMENTS[node.labelModeData.atomicNumber].symbol}${hydrogensLabel}`;
            node.focusCharacterRange = [0, ELEMENTS[node.labelModeData.atomicNumber].symbol.length];
            node.focusLine = 0;
          }
        } else {
          // There are no hydrogens.

          //Editing the node.
          node.label = ELEMENTS[node.labelModeData.atomicNumber].symbol;
          node.focusCharacterRange = [0, ELEMENTS[node.labelModeData.atomicNumber].symbol.length];
          node.focusLine = 0;
        }
      }
    }
  });
}
