import { ELEMENTS } from './elements';
import positionFloatingHydrogens from './positionFloatingHydrogens';

// returns a fragment
export default (f) => {


	// const div = document.createElement('DIV');
	// div.innerHTML = f;


  const parser = new DOMParser();
  const div = parser.parseFromString(f, "application/xml").documentElement;

  const nodes = div.getElementsByTagName('n');

	// f = f.split(/\r?\n/);
	//
	// const countsLine = f[3];
	//
	// const numberOfAtoms = Number(countsLine.substring(0,3));
	// const numberOfBonds = Number(countsLine.substring(3,6));
	//
  const fragment = {
    nodes: Array.prototype.map.call(nodes, node => {

      const p = node.getAttribute('p');
      let x = 0, y = 0;
      if (p) {
        const pSplit = p.split(' ');
        x = Number(pSplit[0]);
        y = Number(pSplit[1]);
      }

      // const label = node.textContent;

      let atomicNumber = 0;
      if (node.hasAttribute('Element')) {
        const atomicNumberCandidate = Number(node.getAttribute('Element'));
        if (ELEMENTS[atomicNumberCandidate]) {
          atomicNumber = atomicNumberCandidate;
        }
      } else {
        atomicNumber = 6;
      }

      const numberOfHydrogens = Number(node.getAttribute('NumHydrogens'));

      return {
          x: Math.floor(x * 2 + 49999),
          y: Math.floor(y * 2 + 49999),
          z: 49999, //: Math.floor(Number(atomLine.substring(20,30)) * 50),
          label: (atomicNumber === 6 ? '' : `${ELEMENTS[atomicNumber].symbol}${numberOfHydrogens ? `H${numberOfHydrogens === 1 ? '' : numberOfHydrogens}` : ''}`),
          focusCharacterRange: (atomicNumber === 6 ? [0, 0] : [0, ELEMENTS[atomicNumber].symbol.length]), //[0, label.length], //: [0, atomLine.substring(31,34).trim().length], //[1, 1+atomLine.substring(31,34).trim().length], // one range per line
          focusLine: 0, // assuming one line
          // Floating hydrogens mean that in the -CH2 label, "H2" part occupies
          // the most optimal position (either to the left, above, to the right,
          // or below) so that the least amount of bonds are hindered.
          // labelMode can be:
          // - CHEMICAL_ELEMENT_PLUS_FLOATING_HYDROGENS
          // - MANUAL
          labelMode: 'CHEMICAL_ELEMENT_PLUS_FLOATING_HYDROGENS',
          // If labelMode is MANUAL, then the labelModeData object has no keys
          // and label, focusCharacterRange and focusLine are defined.
          // if labelMode is CHEMICAL_ELEMENT_PLUS_FLOATING_HYDROGENS, the
          // labelModeData has the atomicNumber and numberOfHydrogens keys
          // and based on those, we set and update label, focusCharacterRange and focusLine.
          labelModeData: {
            atomicNumber, // is a JavaScript number
            numberOfHydrogens
          },
          beingEdited: false,
          latestMouseDownXY: {
            x: undefined,
            y: undefined
          }
        };
    }),
    bonds: Array.prototype.map.call(div.getElementsByTagName('b'), bond => {

      const firstNodeId = bond.getAttribute('B');
      const secondNodeId = bond.getAttribute('E');


      const firstNodeIndex = Array.prototype.findIndex.call(nodes, node => node.getAttribute('id') === firstNodeId);
      const secondNodeIndex = Array.prototype.findIndex.call(nodes, node => node.getAttribute('id') === secondNodeId);


      return {
          firstNodeIndex,
          secondNodeIndex,
          type: 1,
          stereo: 1
        };
    })
  };
	//
	// // reading the atom block
	// for (let i = 4; i < 4 + numberOfAtoms; ++i) {
	// 	const atomLine = f[i];
	//
  //   const node = {
  //     x: Math.floor(Number(atomLine.substring(0,10)) * 50),
  //     y: Math.floor(- Number(atomLine.substring(10,20)) * 50),
  //     z: Math.floor(Number(atomLine.substring(20,30)) * 50),
  //     label: atomLine.substring(31,34).trim()+'H₂₂',
  //     focusCharacterRange: [0, atomLine.substring(31,34).trim().length], //[1, 1+atomLine.substring(31,34).trim().length], // one range per line
  //     focusLine: 0, // assuming one line
  //     beingEdited: false,
  //     latestMouseDownXY: {
  //       x: undefined,
  //       y: undefined
  //     }
  //   };
	//
  //   fragment.nodes.push(node);
	// }
	//
	// // reading the bond block
	// for (let i = 4 + numberOfAtoms; i < 4 + numberOfAtoms + numberOfBonds; ++i) {
	// 	const bondLine = f[i];
	//
  //   const bond = {
  //     firstNodeIndex: Number(bondLine.substring(0,3)) - 1,
  //     secondNodeIndex: Number(bondLine.substring(3,6)) - 1,
  //     type: Number(bondLine.substring(6,9)),
  //     stereo: Number(bondLine.substring(9,12))
  //   };
	//
  //   fragment.bonds.push(bond);
	// }

  positionFloatingHydrogens(fragment);

  return fragment;
}
